import math
from functools import reduce
from itertools import starmap
from operator import mul

from cubed.types import T_DType, T_RegularChunks, T_Shape
from cubed.utils import itemsize, normalize_dtype


class ArrayMetadata:
    def __init__(
        self,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
    ):
        self.shape = shape
        self.dtype = normalize_dtype(dtype)
        self.chunks = chunks

    @property
    def size(self) -> int:
        """Number of elements in the array."""
        return reduce(mul, self.shape, 1)

    @property
    def nbytes(self) -> int:
        """Number of bytes in array"""
        return self.size * itemsize(self.dtype)

    @property
    def _cdata_shape(self) -> T_Shape:
        """The shape of the chunk grid for this array."""
        return tuple(
            starmap(
                lambda s, c: math.ceil(s / c),
                zip(self.shape, self.chunks, strict=False),
            )
        )

    @property
    def nchunks(self) -> int:
        """Number of chunks in array"""
        return reduce(mul, self._cdata_shape, 1)



import math

from cubed.array_api.data_type_functions import isdtype
from cubed.array_api.dtypes import (
    _floating_dtypes,
    _real_floating_dtypes,
    _real_numeric_dtypes,
    _upcast_integral_dtypes,
)
from cubed.array_api.elementwise_functions import sqrt
from cubed.backend_array_api import namespace as nxp
from cubed.core import reduction, scan


def cumulative_prod(x, /, *, axis=None, dtype=None, include_initial=False, device=None):
    if include_initial:
        raise NotImplementedError("include_initial is not supported in cumulative_prod")
    dtype = _upcast_integral_dtypes(
        x,
        dtype,
        allowed_dtypes=(
            "numeric",
            "boolean",
        ),
        fname="cumulative_prod",
        device=device,
    )
    return scan(
        x,
        preop=nxp.prod,
        func=_cumulative_prod_func,
        binop=nxp.multiply,
        axis=axis,
        dtype=dtype,
    )


def _cumulative_prod_func(a, /, *, axis=None, dtype=None, include_initial=False):
    out = nxp.cumulative_prod(
        a, axis=axis, dtype=dtype, include_initial=include_initial
    )
    if include_initial:
        # we don't yet support including the final element as it complicates chunk sizing
        ind = tuple(
            slice(a.shape[i]) if i == axis else slice(None) for i in range(a.ndim)
        )
        out = out[ind]
    return out


def cumulative_sum(x, /, *, axis=None, dtype=None, include_initial=False, device=None):
    if include_initial:
        raise NotImplementedError("include_initial is not supported in cumulative_sum")
    dtype = _upcast_integral_dtypes(
        x,
        dtype,
        allowed_dtypes=(
            "numeric",
            "boolean",
        ),
        fname="cumulative_sum",
        device=device,
    )
    return scan(
        x,
        preop=nxp.sum,
        func=_cumulative_sum_func,
        binop=nxp.add,
        axis=axis,
        dtype=dtype,
    )


def _cumulative_sum_func(a, /, *, axis=None, dtype=None, include_initial=False):
    out = nxp.cumulative_sum(a, axis=axis, dtype=dtype, include_initial=include_initial)
    if include_initial:
        # we don't yet support including the final element as it complicates chunk sizing
        ind = tuple(
            slice(a.shape[i]) if i == axis else slice(None) for i in range(a.ndim)
        )
        out = out[ind]
    return out


def max(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in max")
    return reduction(
        x,
        nxp.max,
        axis=axis,
        dtype=x.dtype,
        split_every=split_every,
        keepdims=keepdims,
    )


def mean(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in mean")
    # This implementation uses a Zarr group of two arrays to store a
    # pair of fields needed to keep per-chunk counts and totals for computing
    # the mean.
    dtype = x.dtype
    # TODO(#658): Should these be default dtypes?
    if isdtype(x.dtype, "complex floating"):
        intermediate_dtype = [("n", nxp.int64), ("total", nxp.complex128)]
    else:
        intermediate_dtype = [("n", nxp.int64), ("total", nxp.float64)]
    extra_func_kwargs = dict(dtype=intermediate_dtype)
    return reduction(
        x,
        _mean_func,
        combine_func=_mean_combine,
        aggregate_func=_mean_aggregate,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
        extra_func_kwargs=extra_func_kwargs,
    )


def _mean_func(a, **kwargs):
    dtype = dict(kwargs.pop("dtype"))
    n = _numel(a, dtype=dtype["n"], **kwargs)
    total = nxp.sum(a, dtype=dtype["total"], **kwargs)
    return {"n": n, "total": total}


def _mean_combine(a, **kwargs):
    dtype = dict(kwargs.pop("dtype"))
    n = nxp.sum(a["n"], dtype=dtype["n"], **kwargs)
    total = nxp.sum(a["total"], dtype=dtype["total"], **kwargs)
    return {"n": n, "total": total}


def _mean_aggregate(a, **kwargs):
    return nxp.divide(a["total"], a["n"])


# based on dask
def _numel(x, **kwargs):
    """
    A reduction to count the number of elements.
    """
    shape = x.shape
    keepdims = kwargs.get("keepdims", False)
    axis = kwargs.get("axis", None)
    dtype = kwargs.get("dtype", nxp.float64)

    if axis is None:
        prod = nxp.prod(shape, dtype=dtype)
        if keepdims is False:
            return prod

        return nxp.full(shape=(1,) * len(shape), fill_value=prod, dtype=dtype)

    if not isinstance(axis, (tuple, list)):
        axis = [axis]

    prod = math.prod(shape[dim] for dim in axis)
    if keepdims is True:
        new_shape = tuple(
            shape[dim] if dim not in axis else 1 for dim in range(len(shape))
        )
    else:
        new_shape = tuple(shape[dim] for dim in range(len(shape)) if dim not in axis)

    return nxp.broadcast_to(nxp.asarray(prod, dtype=dtype), new_shape)


def min(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in min")
    return reduction(
        x,
        nxp.min,
        axis=axis,
        dtype=x.dtype,
        split_every=split_every,
        keepdims=keepdims,
    )


def prod(x, /, *, axis=None, dtype=None, keepdims=False, split_every=None, device=None):
    dtype = _upcast_integral_dtypes(
        x,
        dtype,
        allowed_dtypes=(
            "numeric",
            "boolean",
        ),
        fname="prod",
        device=device,
    )
    extra_func_kwargs = dict(dtype=dtype)
    return reduction(
        x,
        nxp.prod,
        axis=axis,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
        extra_func_kwargs=extra_func_kwargs,
    )


def std(x, /, *, axis=None, correction=0.0, keepdims=False, split_every=None):
    return sqrt(
        var(
            x,
            axis=axis,
            correction=correction,
            keepdims=keepdims,
            split_every=split_every,
        )
    )


def sum(x, /, *, axis=None, dtype=None, keepdims=False, split_every=None, device=None):
    dtype = _upcast_integral_dtypes(
        x,
        dtype,
        allowed_dtypes=(
            "numeric",
            "boolean",
        ),
        fname="sum",
        device=device,
    )
    extra_func_kwargs = dict(dtype=dtype)
    return reduction(
        x,
        nxp.sum,
        axis=axis,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
        extra_func_kwargs=extra_func_kwargs,
    )


def var(
    x,
    /,
    *,
    axis=None,
    correction=0.0,
    keepdims=False,
    split_every=None,
):
    # This implementation follows https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

    if x.dtype not in _real_floating_dtypes:
        raise TypeError("Only real floating-point dtypes are allowed in var")
    dtype = x.dtype
    # TODO(#658): Should these be default dtypes?
    intermediate_dtype = [("n", nxp.int64), ("mu", nxp.float64), ("M2", nxp.float64)]
    extra_func_kwargs = dict(dtype=intermediate_dtype, correction=correction)
    return reduction(
        x,
        _var_func,
        combine_func=_var_combine,
        aggregate_func=_var_aggregate,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
        extra_func_kwargs=extra_func_kwargs,
    )


def _var_func(a, correction=None, **kwargs):
    dtype = dict(kwargs.pop("dtype"))
    n = _numel(a, dtype=dtype["n"], **kwargs)
    mu = nxp.mean(a, dtype=dtype["mu"], **kwargs)
    M2 = nxp.sum(nxp.square(a - mu), dtype=dtype["M2"], **kwargs)
    return {"n": n, "mu": mu, "M2": M2}


def _var_combine(a, axis=None, correction=None, **kwargs):
    # _var_combine is called by _partial_reduce which concatenates along the first axis
    axis = axis[0]
    if a["n"].shape[axis] == 1:  # nothing to combine
        return a
    if a["n"].shape[axis] != 2:
        raise ValueError(f"Expected two elements in {axis} axis to combine")

    n_a = nxp.take(a["n"], 0, axis=axis)
    n_b = nxp.take(a["n"], 1, axis=axis)
    mu_a = nxp.take(a["mu"], 0, axis=axis)
    mu_b = nxp.take(a["mu"], 1, axis=axis)
    M2_a = nxp.take(a["M2"], 0, axis=axis)
    M2_b = nxp.take(a["M2"], 1, axis=axis)

    n_ab = n_a + n_b
    delta = mu_b - mu_a
    mu_ab = (n_a * mu_a + n_b * mu_b) / n_ab
    M2_ab = M2_a + M2_b + delta**2 * n_a * n_b / n_ab

    n = nxp.expand_dims(n_ab, axis=axis)
    mu = nxp.expand_dims(mu_ab, axis=axis)
    M2 = nxp.expand_dims(M2_ab, axis=axis)

    return {"n": n, "mu": mu, "M2": M2}


def _var_aggregate(a, correction=None, **kwargs):
    return nxp.divide(a["M2"], a["n"] - correction)



from dataclasses import asdict
from pathlib import Path
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback

matplotlib.use("Agg")


class MemoryVisualizationCallback(Callback):
    def __init__(self, format: Optional[str] = "svg") -> None:
        self.format = format

    def on_compute_start(self, event):
        plan = []
        for name, node in visit_nodes(event.dag):
            primitive_op = node["primitive_op"]
            plan.append(
                dict(
                    name=name,
                    op_name=node["op_name"],
                    projected_mem=primitive_op.projected_mem,
                    allowed_mem=primitive_op.allowed_mem,
                    reserved_mem=primitive_op.reserved_mem,
                    num_tasks=primitive_op.num_tasks,
                )
            )

        self.plan = plan
        self.events = []

    def on_task_end(self, event):
        self.events.append(asdict(event))

    def on_compute_end(self, event):
        events_df = pd.DataFrame(self.events)
        plan_df = pd.DataFrame(self.plan)
        fig = generate_mem_usage(events_df, plan_df)

        self.dst = Path(f"history/{event.compute_id}")
        self.dst.mkdir(parents=True, exist_ok=True)
        self.dst = self.dst / f"memory.{self.format}"

        fig.savefig(self.dst)


def generate_mem_usage(events_df, plan_df):
    # colours match those in https://cubed-dev.github.io/cubed/user-guide/memory.html

    events_df = events_df.sort_values(by=["task_create_tstamp", "name"], ascending=True)
    projected_mem_map = plan_df.set_index("name")["projected_mem"].to_dict()

    tstamp = events_df["task_result_tstamp"].astype("timedelta64[s]")
    events_df["time"] = (tstamp - tstamp.min()).astype(int)
    events_df["actual usage"] = events_df["peak_measured_mem_end"] / 1_000_000
    events_df["projected_mem"] = events_df.name.map(projected_mem_map) / 1_000_000

    fig, ax = plt.subplots(figsize=(8, 6))

    events_df.plot(
        kind="area", y="actual usage", ax=ax, use_index=True, color="#9fc5e8"
    )

    allowed_mem = plan_df["allowed_mem"].max() / 1_000_000
    ax.axhline(allowed_mem, label="allowed", color="#e06666", linestyle="--")

    reserved_mem = plan_df["reserved_mem"].max() / 1_000_000
    ax.axhline(
        reserved_mem,
        label="reserved",
        color="#f6b26b",
        linestyle="--",
    )

    peak_measured_mem = events_df["peak_measured_mem_end"].max() / 1_000_000
    ax.axhline(peak_measured_mem, label="max actual usage", color="#6fa8dc")

    events_df.plot(
        kind="line",
        y="projected_mem",
        ax=ax,
        use_index=True,
        label="projected",
        color="#93c47d",
        linestyle="--",
    )

    ax.set_xlabel("Task number")
    ax.set_ylim(top=allowed_mem + 100)
    ax.set_ylabel("Task memory (MB)")
    ax.legend()

    return fig



import numpy as np
import pytest

import cubed
import cubed.array_api as xp
from cubed._testing import assert_array_equal


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


# Int array indexing is not a part of the array API, so test it separately.


@pytest.mark.parametrize(
    "ind",
    [
        [6, 7, 2, 9, 10],
        ([6, 7, 2, 9, 10], xp.newaxis),
        (xp.newaxis, [6, 7, 2, 9, 10]),
        (slice(None), xp.newaxis),
        (xp.newaxis, slice(None)),
    ],
)
def test_int_array_index_1d(spec, ind):
    a = xp.arange(12, chunks=(3,), spec=spec)
    b = a.rechunk((4,))  # force materialization to test indexing against zarr
    assert_array_equal(b[ind].compute(), np.arange(12)[ind])


@pytest.mark.parametrize(
    "ind",
    [
        (slice(None), [2, 1]),
        ([1, 2, 1], slice(None)),
        (xp.newaxis, slice(None), [2, 1]),
        (slice(None), xp.newaxis, [2, 1]),
        (slice(None), [2, 1], xp.newaxis),
        (xp.newaxis, slice(None), slice(None)),
        (slice(None), xp.newaxis, slice(None)),
        (slice(None), slice(None), xp.newaxis),
    ],
)
def test_int_array_index_2d(spec, ind):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(3, 3),
        spec=spec,
    )
    b = a.rechunk((2, 2))  # force materialization to test indexing against zarr
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    assert_array_equal(b[ind].compute(), x[ind])


@pytest.mark.parametrize(
    "ind",
    [
        (slice(None), [0, 1, 2, 3]),
        ([0, 1, 2, 3], slice(None)),
    ],
)
def test_int_array_index_2d_no_op(spec, ind):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(3, 3),
        spec=spec,
    )
    assert a is a[ind]


def test_multiple_int_array_indexes(spec):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    with pytest.raises(NotImplementedError):
        a[[1, 2, 1], [2, 1, 0]]


def test_blocks():
    # based on dask tests
    x = xp.arange(10, chunks=2)
    assert x.blocks.shape == (5,)
    assert x.blocks.size == 5

    assert_array_equal(x.blocks[0], x[:2])
    assert_array_equal(x.blocks[-1], x[-2:])
    assert_array_equal(x.blocks[:3], x[:6])
    assert_array_equal(x.blocks[[0, 1, 2]], x[:6])
    assert_array_equal(x.blocks[[3, 0, 2]], np.array([6, 7, 0, 1, 4, 5]))

    x = cubed.random.random((20, 20), chunks=(4, 5))
    assert x.blocks.shape == (5, 4)
    assert x.blocks.size == 20
    assert_array_equal(x.blocks[0], x[:4])
    assert_array_equal(x.blocks[0, :3], x[:4, :15])
    assert_array_equal(x.blocks[:, :3], x[:, :15])

    x = xp.ones((40, 40, 40), chunks=(10, 10, 10))
    assert_array_equal(x.blocks[0, :, 0], np.ones((10, 40, 10)))

    x = xp.ones((2, 2), chunks=1)
    with pytest.raises(ValueError, match="newaxis is not supported"):
        x.blocks[xp.newaxis, :, :]
    with pytest.raises(NotImplementedError):
        x.blocks[[0, 1], [0, 1]]
    with pytest.raises(IndexError, match="out of bounds"):
        x.blocks[100, 100]



import math
from typing import Dict, Optional, TypeVar

from cubed.spec import Spec
from cubed.utils import is_cloud_storage_path

T = TypeVar("T")


def use_backups_default(spec: Optional[Spec]) -> bool:
    """The default setting for ``use_backups``.

    Backup tasks are only enabled on cloud object stores since they provide atomic writes.
    """
    return (
        spec is not None
        and spec.work_dir is not None
        and is_cloud_storage_path(spec.work_dir)
    )


def should_launch_backup(
    task: T,
    now: float,
    start_times: Dict[T, float],
    end_times: Dict[T, float],
    min_tasks: int = 10,
    min_completed_fraction: float = 0.5,
    slow_factor: float = 3.0,
) -> bool:
    """
    Determine whether to launch a backup task.

    Backup tasks are only launched if there are at least `min_tasks` being run, and `min_completed_fraction` of tasks have completed.
    If both those criteria have been met, then a backup task is launched if the duration of the current task is at least
    `slow_factor` times slower than the `min_completed_fraction` percentile task duration.
    """
    if len(start_times) < min_tasks:
        return False
    n = math.ceil(len(start_times) * min_completed_fraction) - 1
    if len(end_times) <= n:
        return False
    completed_durations = sorted(
        [end_times[task] - start_times[task] for task in end_times]
    )
    duration = now - start_times[task]
    result = duration > completed_durations[n] * slow_factor
    if result:
        print(
            "should_launch_backup is True, with duration=%s, completed_durations=%s, n=%s, completed_durations[n]=%s, slow_factor=%s"
            % (duration, completed_durations, n, completed_durations[n], slow_factor)
        )
    return result



import math
from operator import add
from typing import TYPE_CHECKING

import ndindex
import numpy as np
from toolz import accumulate, map

from cubed.backend_array_api import backend_array_to_numpy_array
from cubed.core.array import CoreArray
from cubed.core.ops import general_blockwise, map_selection, merge_chunks
from cubed.utils import array_size, normalize_chunks

if TYPE_CHECKING:
    from cubed.array_api.array_object import Array


def index(x, key):
    "Subset an array, along one or more axes."
    if not isinstance(key, tuple):
        key = (key,)

    # Replace Cubed arrays with NumPy arrays - note that this may trigger a computation!
    # Note that NumPy arrays are needed for ndindex.
    key = tuple(
        backend_array_to_numpy_array(dim_sel.compute())
        if isinstance(dim_sel, CoreArray)
        else dim_sel
        for dim_sel in key
    )

    # Canonicalize index
    idx = ndindex.ndindex(key)
    idx = idx.expand(x.shape)

    # Remove newaxis values, to be filled in with expand_dims at end
    where_newaxis = [
        i for i, ia in enumerate(idx.args) if isinstance(ia, ndindex.Newaxis)
    ]
    for i, a in enumerate(where_newaxis):
        n = sum(isinstance(ia, ndindex.Integer) for ia in idx.args[:a])
        if n:
            where_newaxis[i] -= n
    idx = ndindex.Tuple(*(ia for ia in idx.args if not isinstance(ia, ndindex.Newaxis)))
    selection = idx.raw

    # Check selection is supported
    if any(ia.step < 1 for ia in idx.args if isinstance(ia, ndindex.Slice)):
        raise NotImplementedError(f"Slice step must be >= 1: {key}")
    if not all(
        isinstance(ia, (ndindex.Integer, ndindex.Slice, ndindex.IntegerArray))
        for ia in idx.args
    ):
        raise NotImplementedError(
            "Only integer, slice, or integer array indexes are allowed."
        )
    if sum(1 for ia in idx.args if isinstance(ia, ndindex.IntegerArray)) > 1:
        raise NotImplementedError("Only one integer array index is allowed.")

    # Use ndindex to find the resulting array shape and chunks

    def chunk_len_for_indexer(ia, c):
        if not isinstance(ia, ndindex.Slice):
            return c
        return max(c // ia.step, 1)

    def merged_chunk_len_for_indexer(ia, c):
        if not isinstance(ia, ndindex.Slice):
            return c
        if ia.step == 1:
            return c
        if (c // ia.step) < 1:
            return c
        # note that this may not be the same as c
        # but it is guaranteed to be a multiple of the corresponding
        # value returned by chunk_len_for_indexer, which is required
        # by merge_chunks
        return (c // ia.step) * ia.step

    shape = idx.newshape(x.shape)

    if shape == x.shape:
        # no op case (except possibly newaxis applied below)
        out = x
    elif array_size(shape) == 0:
        # empty output case
        from cubed.array_api.creation_functions import empty

        chunks = tuple(c for c in x.chunksize if c > 0)
        out = empty(shape, dtype=x.dtype, chunks=chunks, spec=x.spec)
    else:
        dtype = x.dtype
        chunks = tuple(
            chunk_len_for_indexer(ia, c)
            for ia, c in zip(idx.args, x.chunksize)
            if not isinstance(ia, ndindex.Integer)
        )

        # this is the same as chunks, except it has the same number of dimensions as the input
        out_chunksizes = tuple(
            chunk_len_for_indexer(ia, c) if not isinstance(ia, ndindex.Integer) else 1
            for ia, c in zip(idx.args, x.chunksize)
        )

        target_chunks = normalize_chunks(chunks, shape, dtype=dtype)

        # use map_selection (which uses general_blockwise) to allow more opportunities for optimization than map_direct

        def selection_function(out_key):
            out_coords = out_key[1:]
            return _target_chunk_selection(target_chunks, out_coords, selection)

        max_num_input_blocks = _index_num_input_blocks(
            idx, x.chunksize, out_chunksizes, x.numblocks
        )

        out = map_selection(
            None,  # no function to apply after selection
            selection_function,
            x,
            shape,
            x.dtype,
            target_chunks,
            max_num_input_blocks=max_num_input_blocks,
        )

        # merge chunks for any dims with step > 1 so they are
        # the same size as the input (or slightly smaller due to rounding)
        merged_chunks = tuple(
            merged_chunk_len_for_indexer(ia, c)
            for ia, c in zip(idx.args, x.chunksize)
            if not isinstance(ia, ndindex.Integer)
        )
        if chunks != merged_chunks:
            out = merge_chunks(out, merged_chunks)

    for axis in where_newaxis:
        from cubed.array_api.manipulation_functions import expand_dims

        out = expand_dims(out, axis=axis)

    return out


def _index_num_input_blocks(
    idx: ndindex.Tuple, in_chunksizes, out_chunksizes, numblocks
):
    num = 1
    for ia, c, oc, nb in zip(idx.args, in_chunksizes, out_chunksizes, numblocks):
        if isinstance(ia, ndindex.Integer) or nb == 1:
            pass  # single block
        elif isinstance(ia, ndindex.Slice):
            if (ia.start // c) == ((ia.stop - 1) // c):
                pass  # within same block
            elif ia.start % c != 0:
                num *= 2  # doesn't start on chunk boundary
            elif ia.step is not None and c % ia.step != 0 and oc > 1:
                # step is not a multiple of chunk size, and output chunks have more than one element
                # so some output chunks will access two input chunks
                num *= 2
        elif isinstance(ia, ndindex.IntegerArray):
            # in the worse case, elements could be retrieved from all blocks
            # TODO: improve to calculate the actual max input blocks
            num *= nb
        else:
            raise NotImplementedError(
                "Only integer, slice, or int array indexes are supported."
            )
    return num


def _target_chunk_selection(target_chunks, idx, selection):
    # integer, integer array, and slice indexes can be interspersed in selection
    # idx is the chunk index for the output (target_chunks)

    sel = []
    i = 0  # index into target_chunks and idx
    for s in selection:
        if isinstance(s, slice):
            offset = s.start or 0
            step = s.step if s.step is not None else 1
            start = tuple(
                accumulate(add, tuple(x * step for x in target_chunks[i]), offset)
            )
            j = idx[i]
            sel.append(slice(start[j], start[j + 1], step))
            i += 1
        # ndindex uses np.ndarray for integer arrays
        elif isinstance(s, np.ndarray):
            # find the cumulative chunk starts
            target_chunk_starts = [0] + list(
                accumulate(add, [c for c in target_chunks[i]])
            )
            # and use to slice the integer array
            j = idx[i]
            sel.append(s[target_chunk_starts[j] : target_chunk_starts[j + 1]])
            i += 1
        elif isinstance(s, int):
            sel.append(s)
            # don't increment i since integer indexes don't have a dimension in the target
        else:
            raise ValueError(f"Unsupported selection: {s}")
    return tuple(sel)


class BlockView:
    """An array-like interface to the blocks of an array."""

    def __init__(self, array: "Array"):
        self.array = array

    def __getitem__(self, key) -> "Array":
        if not isinstance(key, tuple):
            key = (key,)

        # Canonicalize index
        idx = ndindex.ndindex(key)
        idx = idx.expand(self.array.numblocks)

        if any(isinstance(ia, ndindex.Newaxis) for ia in idx.args):
            raise ValueError("Slicing with xp.newaxis is not supported")

        if sum(1 for ia in idx.args if isinstance(ia, ndindex.IntegerArray)) > 1:
            raise NotImplementedError("Only one integer array index is allowed.")

        # convert Integer to Slice so we don't lose dimensions
        def convert_integer_index_to_slice(ia):
            if isinstance(ia, ndindex.Integer):
                return ndindex.Slice(ia.raw, ia.raw + 1)
            return ia

        idx = ndindex.Tuple(*(convert_integer_index_to_slice(ia) for ia in idx.args))

        chunks = tuple(
            tuple(np.array(ch)[ia].tolist())
            for ia, ch in zip(idx.raw, self.array.chunks)
        )
        shape = tuple(map(sum, chunks))

        identity = lambda a: a

        def get_dim_index(ia, i):
            if isinstance(ia, ndindex.Slice):
                step = ia.step or 1
                return ia.start + (step * i)
            elif isinstance(ia, ndindex.IntegerArray):
                return ia.raw[i]
            else:
                raise NotImplementedError(
                    "Only integer, slice, or int array indexes are supported."
                )

        def key_function(out_key):
            out_coords = out_key[1:]
            in_coords = tuple(
                get_dim_index(ia, bi) for ia, bi in zip(idx.args, out_coords)
            )
            return ((self.array.name, *in_coords),)

        out = general_blockwise(
            identity,
            key_function,
            self.array,
            shapes=[shape],
            dtypes=[self.array.dtype],
            chunkss=[chunks],
        )

        from cubed import Array

        assert isinstance(out, Array)  # single output
        return out

    @property
    def size(self) -> int:
        """
        The total number of blocks in the array.
        """
        return math.prod(self.shape)

    @property
    def shape(self) -> tuple[int, ...]:
        """
        The number of blocks per axis.
        """
        return self.array.numblocks



from cubed.array_api.creation_functions import asarray, zeros_like
from cubed.array_api.data_type_functions import astype, result_type
from cubed.array_api.dtypes import _promote_scalars, _real_numeric_dtypes
from cubed.array_api.manipulation_functions import reshape
from cubed.array_api.statistical_functions import max, sum
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import arg_reduction, blockwise, elemwise


def argmax(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in argmax")
    if axis is None:
        x = reshape(x, (-1,))
        axis = 0
        keepdims = False
    return arg_reduction(
        x,
        nxp.argmax,
        axis=axis,
        keepdims=keepdims,
        split_every=split_every,
    )


def argmin(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in argmin")
    if axis is None:
        x = reshape(x, (-1,))
        axis = 0
        keepdims = False
    return arg_reduction(
        x,
        nxp.argmin,
        axis=axis,
        keepdims=keepdims,
        split_every=split_every,
    )


def count_nonzero(x, /, *, axis=None, keepdims=False, split_every=None):
    dtype = nxp.__array_namespace_info__().default_dtypes(device=x.device)["indexing"]
    return sum(
        astype(x, nxp.bool),
        axis=axis,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
    )


def searchsorted(x1, x2, /, *, side="left", sorter=None):
    if x1.ndim != 1:
        raise ValueError("Input array x1 must be one dimensional")

    if sorter is not None:
        raise NotImplementedError(
            "searchsorted with a sorter argument is not supported"
        )

    # call nxp.searchsorted for each pair of blocks in x1 and v
    dtype = nxp.__array_namespace_info__().default_dtypes(device=x1.device)["indexing"]
    out = blockwise(
        _searchsorted,
        list(range(x2.ndim + 1)),
        x1,
        [0],
        x2,
        list(range(1, x2.ndim + 1)),
        dtype=dtype,
        adjust_chunks={0: 1},  # one row for each block in x1
        side=side,
    )

    # add offsets to take account of the position of each block within the array x1
    x1_chunk_sizes = nxp.asarray((0, *x1.chunks[0]))
    x1_chunk_offsets = nxp.cumulative_sum(x1_chunk_sizes)[:-1]
    x1_chunk_offsets = x1_chunk_offsets[(Ellipsis,) + x2.ndim * (nxp.newaxis,)]
    x1_offsets = asarray(x1_chunk_offsets, chunks=1)
    out = where(out < 0, out, out + x1_offsets)

    # combine the results from each block (of a)
    out = max(out, axis=0)

    # fix up any -1 values
    # TODO: use general_blockwise which has block_id to avoid this
    out = where(out >= 0, out, zeros_like(out))

    return out


def _searchsorted(x, y, side):
    res = nxp.searchsorted(x, y, side=side)
    # 0 is only correct for the first block of a, but blockwise doesn't have a way
    # of telling which block is being operated on (unlike map_blocks),
    # so set all 0 values to a special value and set back at the end of searchsorted
    res = nxp.where(res == 0, -1, res)
    return res[nxp.newaxis, :]


def where(condition, x1, x2, /):
    x1, x2 = _promote_scalars(x1, x2, "where")
    dtype = result_type(x1, x2)
    return elemwise(nxp.where, condition, x1, x2, dtype=dtype)



import collections.abc
from itertools import product
from typing import Iterator

from cubed.primitive.blockwise import (
    BlockwiseSpec,
    fuse_blockwise_specs,
    make_fused_key_function,
)
from cubed.utils import map_nested


def make_map_blocks_key_function(*names):
    def key_function(out_key):
        out_coords = out_key[1:]
        return tuple((name, *out_coords) for name in names)

    return key_function


def make_combine_blocks_list_key_function(name, numblocks, split_every):
    # similar to the key function in partial_reduce
    def key_function(out_key):
        out_coords = out_key[1:]

        # return a tuple with a single item that is a list of input keys to be combined
        in_keys = [
            list(
                range(
                    bi * split_every,
                    min((bi + 1) * split_every, numblocks),
                )
            )
            for bi in out_coords
        ]
        return ([(name,) + tuple(p) for p in product(*in_keys)],)

    return key_function


def make_combine_blocks_iter_key_function(name, numblocks, split_every):
    # similar to the key function in partial_reduce
    def key_function(out_key):
        out_coords = out_key[1:]

        # return a tuple with a single item that is an iterator of input keys to be combined
        in_keys = [
            list(
                range(
                    bi * split_every,
                    min((bi + 1) * split_every, numblocks),
                )
            )
            for bi in out_coords
        ]
        return (iter([(name,) + tuple(p) for p in product(*in_keys)]),)

    return key_function


def negative(x):
    assert isinstance(x, int)
    return -x


def add(x, y):
    assert isinstance(x, int)
    assert isinstance(y, int)
    return x + y


def sum_iter(x):
    assert isinstance(x, Iterator)
    return sum(a for a in x)


def check_key_function(key_function, out_coords, expected_str):
    res = key_function(("out",) + out_coords)

    assert str(tuple(iter_repr_nested(r) for r in res)) == expected_str


def iter_repr_nested(seq):
    # convert nested iterators to lists
    if isinstance(seq, list):
        return [iter_repr_nested(item) for item in seq]
    elif isinstance(seq, Iterator):
        return IteratorWithRepr([iter_repr_nested(item) for item in seq])
    else:
        return seq


class IteratorWithRepr(collections.abc.Iterator):
    def __init__(self, values):
        self.values = values
        self.it = iter(values)

    def __next__(self):
        return next(self.it)

    def __repr__(self):
        return "<" + ", ".join(repr(v) for v in self.values) + ">"

    def __str__(self):
        return "<" + ", ".join(str(v) for v in self.values) + ">"


def make_blockwise_spec(
    key_function,
    function,
    function_nargs=1,
    num_input_blocks=(1,),
    num_output_blocks=(1,),
    iterable_input_blocks=(False,),
):
    return BlockwiseSpec(
        key_function=key_function,
        function=function,
        function_nargs=function_nargs,
        num_input_blocks=num_input_blocks,
        num_output_blocks=num_output_blocks,
        iterable_input_blocks=iterable_input_blocks,
        reads_map={},  # unused
        writes_map={},  # unused
    )


def test_map_blocks_key_function():
    key_function = make_map_blocks_key_function("a")

    check_key_function(key_function, (0,), "(('a', 0),)")
    check_key_function(key_function, (1,), "(('a', 1),)")


def test_map_blocks_multiple_inputs_key_function():
    key_function = make_map_blocks_key_function("a", "b")

    check_key_function(key_function, (0,), "(('a', 0), ('b', 0))")
    check_key_function(key_function, (1,), "(('a', 1), ('b', 1))")


def test_combine_blocks_list_key_function():
    key_function = make_combine_blocks_list_key_function(
        "a", numblocks=5, split_every=2
    )

    check_key_function(key_function, (0,), "([('a', 0), ('a', 1)],)")
    check_key_function(key_function, (1,), "([('a', 2), ('a', 3)],)")
    check_key_function(key_function, (2,), "([('a', 4)],)")


def test_combine_blocks_iter_key_function():
    key_function = make_combine_blocks_iter_key_function(
        "a", numblocks=5, split_every=2
    )

    check_key_function(key_function, (0,), "(<('a', 0), ('a', 1)>,)")
    check_key_function(key_function, (1,), "(<('a', 2), ('a', 3)>,)")
    check_key_function(key_function, (2,), "(<('a', 4)>,)")


def test_fuse_key_function_single_multiple():
    key_function1 = make_map_blocks_key_function("a")
    key_function2 = make_combine_blocks_iter_key_function(
        "b", numblocks=5, split_every=2
    )
    fused_key_function = make_fused_key_function(key_function2, [key_function1], [1])

    check_key_function(fused_key_function, (0,), "([<('a', 0), ('a', 1)>],)")
    check_key_function(fused_key_function, (1,), "([<('a', 2), ('a', 3)>],)")
    check_key_function(fused_key_function, (2,), "([<('a', 4)>],)")


def test_fuse_key_function_multiple_single():
    key_function1 = make_combine_blocks_iter_key_function(
        "a", numblocks=5, split_every=2
    )
    key_function2 = make_map_blocks_key_function("b")
    fused_key_function = make_fused_key_function(key_function2, [key_function1], [1])

    check_key_function(fused_key_function, (0,), "([<('a', 0), ('a', 1)>],)")
    check_key_function(fused_key_function, (1,), "([<('a', 2), ('a', 3)>],)")
    check_key_function(fused_key_function, (2,), "([<('a', 4)>],)")


def test_fuse_key_function_multiple_multiple():
    key_function1 = make_combine_blocks_iter_key_function(
        "a", numblocks=5, split_every=2
    )
    key_function2 = make_combine_blocks_iter_key_function(
        "b", numblocks=3, split_every=2
    )
    fused_key_function = make_fused_key_function(key_function2, [key_function1], [1])

    check_key_function(
        fused_key_function, (0,), "([<<('a', 0), ('a', 1)>, <('a', 2), ('a', 3)>>],)"
    )
    check_key_function(fused_key_function, (1,), "([<<('a', 4)>>],)")


def apply_blockwise(input_data, out_coords, bw_spec):
    args = []
    out_key = ("out",) + tuple(out_coords)  # array name is ignored by key_function
    in_keys = bw_spec.key_function(out_key)
    for in_key in in_keys:
        # just return the (1D) coord as a value
        def get_data(key):
            name = key[0]
            index = key[1]  # 1d index
            return input_data[name][index]

        arg = map_nested(get_data, in_key)
        args.append(arg)
    return bw_spec.function(*args)


def test_apply_blockwise():
    bw_spec = make_blockwise_spec(
        key_function=make_map_blocks_key_function("a"),
        function=negative,
    )

    input_data = {"a": [0, 1, 2, 3, 4]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(5)]
    assert out == [0, -1, -2, -3, -4]


def test_apply_blockwise_multiple_inputs():
    bw_spec = make_blockwise_spec(
        key_function=make_map_blocks_key_function("a", "b"), function=add
    )

    input_data = {"a": [0, 1, 2, 3, 4], "b": [5, 6, 7, 8, 9]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(5)]
    assert out == [5, 7, 9, 11, 13]


def test_apply_blockwise_iterator():
    bw_spec = make_blockwise_spec(
        key_function=make_combine_blocks_iter_key_function(
            "a", numblocks=5, split_every=2
        ),
        function=sum_iter,
    )

    input_data = {"a": [0, 1, 2, 3, 4]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(3)]
    assert out == [1, 5, 4]


def test_apply_blockwise_fused():
    bw_spec1 = make_blockwise_spec(
        key_function=make_map_blocks_key_function("a"), function=negative
    )
    bw_spec2 = make_blockwise_spec(
        key_function=make_map_blocks_key_function("b"), function=negative
    )

    bw_spec = fuse_blockwise_specs(bw_spec2, bw_spec1)

    input_data = {"a": [0, 1, 2, 3, 4]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(5)]
    assert out == [0, 1, 2, 3, 4]


def test_apply_blockwise_fused_iterator():
    bw_spec1 = make_blockwise_spec(
        key_function=make_map_blocks_key_function("a"), function=negative
    )
    bw_spec2 = make_blockwise_spec(
        key_function=make_combine_blocks_iter_key_function(
            "b", numblocks=5, split_every=2
        ),
        function=sum_iter,
        num_input_blocks=(2,),
        iterable_input_blocks=(True,),
    )

    bw_spec = fuse_blockwise_specs(bw_spec2, bw_spec1)

    input_data = {"a": [0, 1, 2, 3, 4]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(3)]
    assert out == [-1, -5, -4]


def test_apply_blockwise_fused_iterator_with_single_input_block():
    bw_spec1 = make_blockwise_spec(
        key_function=make_map_blocks_key_function("a"), function=negative
    )
    # the iterator only reads from a single input block
    bw_spec2 = make_blockwise_spec(
        key_function=make_combine_blocks_iter_key_function(
            "b", numblocks=5, split_every=1
        ),
        function=sum_iter,
        num_input_blocks=(1,),
        iterable_input_blocks=(True,),
    )

    bw_spec = fuse_blockwise_specs(bw_spec2, bw_spec1)

    input_data = {"a": [0, 1, 2, 3, 4]}
    out = [apply_blockwise(input_data, [i], bw_spec) for i in range(5)]
    assert out == [0, -1, -2, -3, -4]



from typing import NamedTuple

from cubed.array_api.array_object import Array
from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import _floating_dtypes

# These functions are in both the main and linalg namespaces
from cubed.array_api.linear_algebra_functions import (  # noqa: F401
    matmul,
    matrix_transpose,
    tensordot,
    vecdot,
)
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import blockwise, general_blockwise, merge_chunks, squeeze
from cubed.utils import array_memory, get_item


def outer(x1, x2, /):
    return blockwise(
        nxp.linalg.outer, "ij", x1, "i", x2, "j", dtype=result_type(x1, x2)
    )


class QRResult(NamedTuple):
    Q: Array
    R: Array


class SVDResult(NamedTuple):
    U: Array
    S: Array
    Vh: Array


def qr(x, /, *, mode="reduced") -> QRResult:
    if x.ndim != 2:
        raise ValueError("qr requires x to have 2 dimensions.")

    if mode != "reduced":
        raise ValueError("qr only supports mode='reduced'")

    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in qr")

    if x.numblocks[1] > 1:
        raise ValueError(
            "qr only supports tall-and-skinny (single column chunk) arrays. "
            "Consider rechunking so there is only a single column chunk."
        )

    Q, R, _, _, _ = tsqr(x)
    return QRResult(Q, R)


def tsqr(x, compute_svd=False, finalize_svd=True):
    """Direct Tall-and-Skinny QR algorithm

    From:

        Direct QR factorizations for tall-and-skinny matrices in MapReduce architectures
        Austin R. Benson, David F. Gleich, James Demmel
        Proceedings of the IEEE International Conference on Big Data, 2013
        https://arxiv.org/abs/1301.1071
    """

    # follows Algorithm 2 from Benson et al, modified for SVD
    Q1, R1 = _qr_first_step(x)

    if _r1_is_too_big(R1):
        R1 = _rechunk_r1(R1)
        Q2, R2, U, S, Vh = tsqr(R1, compute_svd=compute_svd, finalize_svd=False)
    else:
        Q2, R2, U, S, Vh = _qr_second_step(R1, compute_svd=compute_svd)

    Q, R = _qr_third_step(Q1, Q2), R2

    if compute_svd and finalize_svd:
        U = Q @ U  # fourth step (SVD only)
        S = squeeze(S, axis=1)  # remove extra dim

    return Q, R, U, S, Vh


def _qr_first_step(A):
    m, n = A.chunksize
    k, _ = A.numblocks

    # Q1 has same shape and chunks as A
    R1_shape = (n * k, n)
    R1_chunks = ((n,) * k, (n,))
    # qr implementation creates internal array buffers
    extra_projected_mem = A.chunkmem * 4
    Q1, R1 = map_blocks_multiple_outputs(
        nxp.linalg.qr,
        A,
        shapes=[A.shape, R1_shape],
        dtypes=[A.dtype, A.dtype],
        chunkss=[A.chunks, R1_chunks],
        extra_projected_mem=extra_projected_mem,
    )
    return QRResult(Q1, R1)


def _r1_is_too_big(R1):
    array_mem = array_memory(R1.dtype, R1.shape)
    # conservative values for max_mem (4 copies, doubled to give some slack)
    max_mem = (R1.spec.allowed_mem - R1.spec.reserved_mem) // (4 * 2)
    return array_mem > max_mem


def _rechunk_r1(R1, split_every=4):
    # expand R1's chunk size in axis 0 so that new R1 will be smaller by factor of split_every
    if R1.numblocks[0] == 1:
        raise ValueError(
            "Can't expand R1 chunk size further. Try increasing allowed_mem"
        )
    chunks = (R1.chunksize[0] * split_every, R1.chunksize[1])
    return merge_chunks(R1, chunks=chunks)


def _qr_second_step(R1, compute_svd=False):
    R1_single = _merge_into_single_chunk(R1)

    Q2_shape = R1.shape
    Q2_chunks = Q2_shape  # single chunk

    n = R1.shape[1]
    R2_shape = (n, n)
    R2_chunks = R2_shape  # single chunk

    if not compute_svd:
        # qr implementation creates internal array buffers
        extra_projected_mem = R1_single.chunkmem * 4
        Q2, R2 = map_blocks_multiple_outputs(
            nxp.linalg.qr,
            R1_single,
            shapes=[Q2_shape, R2_shape],
            dtypes=[R1.dtype, R1.dtype],
            chunkss=[Q2_chunks, R2_chunks],
            extra_projected_mem=extra_projected_mem,
        )
        return Q2, R2, None, None, None
    else:
        U_shape = (n, n)
        U_chunks = U_shape
        S_shape = (n, 1)  # extra dim since multiple outputs must have same numblocks
        S_chunks = S_shape
        Vh_shape = (n, n)
        Vh_chunks = Vh_shape

        # qr implementation creates internal array buffers
        extra_projected_mem = R1_single.chunkmem * 4
        Q2, R2, U, S, Vh = map_blocks_multiple_outputs(
            _qr2,
            R1_single,
            shapes=[Q2_shape, R2_shape, U_shape, S_shape, Vh_shape],
            dtypes=[R1.dtype, R1.dtype, R1.dtype, R1.dtype, R1.dtype],
            chunkss=[Q2_chunks, R2_chunks, U_chunks, S_chunks, Vh_chunks],
            extra_projected_mem=extra_projected_mem,
        )
        return Q2, R2, U, S, Vh


def _merge_into_single_chunk(x, split_every=4):
    # do a tree merge along first axis
    while x.numblocks[0] > 1:
        chunks = (x.chunksize[0] * split_every,) + x.chunksize[1:]
        x = merge_chunks(x, chunks)
    return x


def _qr2(a):
    Q, R = nxp.linalg.qr(a)
    U, S, Vh = nxp.linalg.svd(R)
    S = S[:, nxp.newaxis]  # add extra dim
    return Q, R, U, S, Vh


def _qr_third_step(Q1, Q2):
    m, n = Q1.chunksize
    k, _ = Q1.numblocks

    Q1_shape = Q1.shape
    Q1_chunks = Q1.chunks

    Q2_single = _merge_into_single_chunk(Q2)

    # These aren't the actual chunks, but the chunks we need for _q_matmul
    Q2_chunks = ((n,) * k, (n,))

    def key_function(out_key):
        # Q1 is a simple 1:1 mapping, Q2_single has a single chunk
        return ((Q1.name,) + out_key[1:], (Q2_single.name,) + (0, 0))

    Q = general_blockwise(
        _q_matmul,
        key_function,
        Q1,
        Q2_single,
        shapes=[Q1_shape],
        dtypes=[result_type(Q1, Q2_single)],
        chunkss=[Q1_chunks],
        q2_chunks=Q2_chunks,
    )
    return Q


def _q_matmul(a1, a2, q2_chunks=None, block_id=None):
    q1 = a1
    # this array only has a single chunk, but we need to get a slice corresponding to q2_chunks
    q2 = a2[get_item(q2_chunks, block_id)]
    return q1 @ q2


def svd(x, /, *, full_matrices=True) -> SVDResult:
    if full_matrices:
        raise ValueError("Cubed arrays only support using full_matrices=False")

    nb = x.numblocks
    # TODO: optimize case nb[0] == nb[1] == 1
    if nb[0] > nb[1]:
        _, _, U, S, Vh = tsqr(x, compute_svd=True)
        truncate = x.shape[0] < x.shape[1]
    else:
        _, _, Vht, S, Ut = tsqr(x.T, compute_svd=True)
        U, S, Vh = Ut.T, S, Vht.T
        truncate = x.shape[0] > x.shape[1]
    if truncate:  # from dask
        k = min(x.shape)
        U, Vh = U[:, :k], Vh[:k, :]
    return SVDResult(U, S, Vh)


def svdvals(x, /):
    _, S, _ = svd(x, full_matrices=False)
    return S


def map_blocks_multiple_outputs(
    func,
    *args,
    shapes,
    dtypes,
    chunkss,
    **kwargs,
):
    def key_function(out_key):
        return tuple((array.name,) + out_key[1:] for array in args)

    return general_blockwise(
        func,
        key_function,
        *args,
        shapes=shapes,
        dtypes=dtypes,
        chunkss=chunkss,
        target_stores=[None] * len(dtypes),
        **kwargs,
    )



import asyncio
import time
from pathlib import Path

import obstore as obs


def path_to_store(path):
    if isinstance(path, str):
        if "://" not in path:
            return obs.store.from_url(Path(path).as_uri(), mkdir=True)
        else:
            return obs.store.from_url(path)
    elif isinstance(path, Path):
        return obs.store.from_url(path.as_uri(), mkdir=True)


def read_int_from_file(store, path):
    result = obs.get(store, path)
    return int(result.bytes())


def write_int_to_file(store, path, i):
    obs.put(store, path, bytes(str(i), encoding="UTF8"))


def deterministic_failure(path, timing_map, i, *, default_sleep=0.01, name=None):
    """A function that can either run normally, run slowly, or raise
    an exception, depending on input and invocation count.

    The timing_map is a dictionary whose keys are inputs and values
    are sequences of timing information for each invocation.
    The maginitude of the value is the time to sleep in seconds, and
    the sign indicates the input is returned normally (positive, or 0),
    or an exception is raised (negative).

    If a input is missing then all invocations will run normally, with a
    small default sleep to avoid spurious backups being launched.

    If there are subsequent invocations to the ones in the sequence, then
    they will all run normally.
    """
    # increment number of invocations of this function with arg i
    store = path_to_store(path)
    try:
        invocation_count = read_int_from_file(store, f"{i}")
    except FileNotFoundError:
        invocation_count = 0
    write_int_to_file(store, f"{i}", invocation_count + 1)

    timing_code = default_sleep
    if i in timing_map:
        timing_codes = timing_map[i]
        if invocation_count >= len(timing_codes):
            timing_code = default_sleep
        else:
            timing_code = timing_codes[invocation_count]

    if timing_code >= 0:
        time.sleep(timing_code)
        return i
    else:
        time.sleep(-timing_code)
        raise RuntimeError(
            f"Deliberately fail on invocation number {invocation_count + 1} for input {i}"
        )


def check_invocation_counts(
    path, timing_map, n_tasks, retries=None, expected_invocation_counts_overrides=None
):
    asyncio.run(
        check_invocation_counts_async(
            path,
            timing_map,
            n_tasks,
            retries=retries,
            expected_invocation_counts_overrides=expected_invocation_counts_overrides,
        )
    )


async def check_invocation_counts_async(
    path, timing_map, n_tasks, retries=None, expected_invocation_counts_overrides=None
):
    expected_invocation_counts = {}
    for i in range(n_tasks):
        if i not in timing_map:
            expected_invocation_counts[i] = 1
        else:
            timing_codes = timing_map[i]
            expected_invocation_count = len(timing_codes) + 1

            if retries is not None:
                # there shouldn't have been more than retries + 1 invocations
                max_invocations = retries + 1
                expected_invocation_count = min(
                    expected_invocation_count, max_invocations
                )

            expected_invocation_counts[i] = expected_invocation_count

    if expected_invocation_counts_overrides is not None:
        expected_invocation_counts.update(expected_invocation_counts_overrides)

    # retrieve outputs concurrently, so we can test on large numbers of inputs
    store = path_to_store(path)
    paths = [str(i) for i in range(n_tasks)]
    results = await asyncio.gather(*[obs.get_async(store, path) for path in paths])
    values = await asyncio.gather(*[result.bytes_async() for result in results])
    actual_invocation_counts = {i: int(val) for i, val in enumerate(values)}

    if actual_invocation_counts != expected_invocation_counts:
        for i, expected_count in expected_invocation_counts.items():
            actual_count = actual_invocation_counts[i]
            if actual_count != expected_count:
                print(
                    f"Invocation count for {i}, expected: {expected_count}, actual: {actual_count}"
                )
    assert actual_invocation_counts == expected_invocation_counts



import pytest

import cubed
import cubed.array_api as xp
from cubed._testing import assert_array_equal
from cubed.storage.backend import backend_storage_name


# This is less strict than the spec, but is supported by implementations like NumPy
def test_prod_sum_bool():
    a = xp.ones((2,), dtype=xp.bool)
    assert_array_equal(xp.prod(a).compute(), xp.asarray([1], dtype=xp.int64))
    assert_array_equal(xp.sum(a).compute(), xp.asarray([2], dtype=xp.int64))


@pytest.mark.skipif(
    backend_storage_name() != "zarr-python",
    reason="object dtype only works on zarr-python",
)
def test_object_dtype():
    a = xp.asarray(["a", "b"], dtype=object, chunks=2)
    cubed.to_zarr(a, store=None)



from typing import Optional

from cubed import config
from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store


def backend_storage_name():
    # get storage name from top-level config
    # e.g. set globally with CUBED_STORAGE_NAME=tensorstore
    storage_name = config.get("storage_name", None)

    if storage_name is None:
        import zarr

        if zarr.__version__[0] == "3":
            storage_name = "zarr-python-v3"
        else:
            storage_name = "zarr-python"

    return storage_name


def is_backend_storage_array(obj):
    storage_name = backend_storage_name()

    if storage_name == "zarr-python":
        import zarr

        from cubed.storage.backends.zarr_python import ZarrArrayGroup

        return isinstance(obj, (zarr.Array, ZarrArrayGroup))
    elif storage_name in ("zarr-python-v3", "zarrs-python"):
        import zarr

        from cubed.storage.backends.zarr_python_v3 import ZarrV3ArrayGroup

        return isinstance(obj, (zarr.Array, ZarrV3ArrayGroup))
    elif storage_name == "tensorstore":
        from cubed.storage.backends.tensorstore import (
            TensorStoreArray,
            TensorStoreGroup,
        )

        return isinstance(obj, (TensorStoreArray, TensorStoreGroup))
    else:
        raise ValueError(f"Unrecognized storage name: {storage_name}")


def open_backend_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    **kwargs,
):
    storage_name = backend_storage_name()

    if storage_name == "zarr-python":
        from cubed.storage.backends.zarr_python import open_zarr_array  # type: ignore

        open_func = open_zarr_array

        # set object codec if needed
        import numpy as np

        if np.dtype(dtype).hasobject and "object_codec" not in kwargs:
            import numcodecs

            object_codec = numcodecs.Pickle()
            kwargs["object_codec"] = object_codec

    elif storage_name == "zarr-python-v3":
        from cubed.storage.backends.zarr_python_v3 import open_zarr_v3_array

        open_func = open_zarr_v3_array
    elif storage_name == "zarrs-python":
        from cubed.storage.backends.zarrs_python import open_zarr_v3_array

        open_func = open_zarr_v3_array
    elif storage_name == "tensorstore":
        from cubed.storage.backends.tensorstore import open_tensorstore_array

        open_func = open_tensorstore_array
    else:
        raise ValueError(f"Unrecognized storage name: {storage_name}")

    return open_func(
        store,
        mode,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        path=path,
        **kwargs,
    )



import pytest

import cubed
import cubed as xp


@pytest.mark.parametrize(
    (
        "min_mem",
        "expected_num_stages",
        "expected_max_input_blocks",
        "expected_max_output_blocks",
    ),
    [
        (None, 3, 16, 15),  # multistage rechunk - more stages, lower fan in/out
        (0, 1, 3771, 3460),  # single stage rechunk - very high fan in/out
    ],
)
def test_rechunk_era5(
    tmp_path,
    min_mem,
    expected_num_stages,
    expected_max_input_blocks,
    expected_max_output_blocks,
):
    # from https://github.com/pangeo-data/rechunker/pull/89
    shape = (350640, 721, 1440)
    source_chunks = (31, 721, 1440)
    target_chunks = (350640, 10, 10)

    spec = cubed.Spec(allowed_mem="2.5GB")

    a = xp.empty(shape, dtype=xp.float32, chunks=source_chunks, spec=spec)
    b = a.rechunk(target_chunks, min_mem=min_mem, use_new_impl=True)

    b.visualize(filename=tmp_path / "rechunk_era5")

    assert b.shape == a.shape

    # find rechunk ops in plan
    plan = b.plan._finalize()
    rechunks = [
        (n, d)
        for (n, d) in plan.dag.nodes(data=True)
        if d.get("op_name", None) == "rechunk"
    ]

    # each stage has two ops due to intermediate store
    assert len(rechunks) == expected_num_stages * 2

    max_input_blocks = max(
        d["pipeline"].config.num_input_blocks[0] for _, d in rechunks
    )
    assert max_input_blocks == expected_max_input_blocks

    max_output_blocks = max(
        d["pipeline"].config.num_output_blocks[0] for _, d in rechunks
    )
    assert max_output_blocks == expected_max_output_blocks


@pytest.mark.parametrize(
    "spec",
    [
        cubed.Spec(allowed_mem="2.5GB"),
        # cloud stores use extra buffer copies, so need more memory for same rechunk plan
        cubed.Spec("s3://cubed-unittest/rechunk-era5", allowed_mem="3.5GB"),
    ],
)
def test_rechunk_era5_chunk_sizes(spec):
    # from https://github.com/pangeo-data/rechunker/pull/89
    shape = (350640, 721, 1440)
    source_chunks = (31, 721, 1440)
    target_chunks = (350640, 10, 10)

    a = xp.empty(shape, dtype=xp.float32, chunks=source_chunks, spec=spec)

    from cubed.core.ops import _rechunk_plan

    assert list(_rechunk_plan(a, target_chunks)) == [
        ((93, 721, 1440), (93, 173, 396)),
        ((1447, 173, 396), (1447, 173, 396)),
        ((1447, 173, 396), (1447, 41, 109)),
        ((22528, 41, 109), (22528, 41, 109)),
        ((22528, 41, 109), (22528, 10, 30)),
        ((350640, 10, 30), (350640, 10, 10)),
    ]



import math

try:
    from math import lcm  # type: ignore  # Python 3.9
except ImportError:

    def lcm(a: int, b: int) -> int:  # type: ignore
        """Implementation of `math.lcm()` for all Python versions."""
        # https://stackoverflow.com/a/51716959/809705
        return a * b // math.gcd(a, b)



import inspect
import itertools
import logging
import math
from collections.abc import Iterator
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import toolz
import zarr
from toolz import map

from cubed.backend_array_api import (
    backend_array_to_numpy_array,
    numpy_array_to_backend_array,
)
from cubed.primitive.memory import BufferCopies, MemoryModeller, calculate_projected_mem
from cubed.primitive.types import CubedArrayProxy, PrimitiveOperation
from cubed.runtime.types import CubedPipeline
from cubed.storage.backend import is_backend_storage_array
from cubed.storage.zarr import LazyZarrArray, T_ZarrArray, lazy_zarr_array
from cubed.types import T_Chunks, T_DType, T_RegularChunks, T_Shape, T_Store
from cubed.utils import (
    array_memory,
    chunk_memory,
    get_item,
    map_nested,
    normalize_chunks,
    split_into,
    to_chunksize,
)
from cubed.utils import numblocks as compute_numblocks
from cubed.vendor.dask.blockwise import _get_coord_mapping, _make_dims, lol_product
from cubed.vendor.dask.core import flatten

logger = logging.getLogger(__name__)


sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


@dataclass(frozen=True)
class BlockwiseSpec:
    """Specification for how to run blockwise on an array.

    This is similar to ``CopySpec`` in rechunker.

    Attributes
    ----------
    key_function : Callable
        A function that maps an output chunk key to one or more input chunk keys.
    function : Callable
        A function that maps input chunks to an output chunk.
    function_nargs: int
        The number of array arguments that ``function`` takes. Note that for some
        functions (``concat``, ``stack``) this may be different to the number of
        arrays that the operation depends on.
    num_input_blocks: Tuple[int, ...]
        The number of input blocks read from each input array.
    num_output_blocks: Tuple[int, ...]
        The number of output blocks written to each output array.
    iterable_input_blocks: Tuple[int, ...]
        Whether the input blocks read from each input array are supplied as an iterable or not.
    reads_map : Dict[str, CubedArrayProxy]
        Read proxy dictionary keyed by array name.
    writes_map : Dict[str, CubedArrayProxy]
        Write proxy dictionary keyed by array name where entries have an ``array`` attribute that supports ``__setitem__``.
        Note that the order of the entries must correspond to the order of the outputs created by the function.
    """

    key_function: Callable[..., Any]
    function: Callable[..., Any]
    function_nargs: int
    num_input_blocks: Tuple[int, ...]
    num_output_blocks: Tuple[int, ...]
    iterable_input_blocks: Tuple[bool, ...]
    reads_map: Dict[str, CubedArrayProxy]
    writes_map: Dict[str, CubedArrayProxy]
    return_writes_stores: bool = False


def apply_blockwise(
    out_coords: List[int], *, config: BlockwiseSpec
) -> Optional[List[T_Store]]:
    """Stage function for blockwise."""
    # lithops needs params to be lists not tuples, so convert back
    out_coords_tuple = tuple(out_coords)

    results = get_results_in_different_scope(out_coords, config=config)

    # if blockwise function is a regular function (not a generator) that doesn't return multiple values then make it iterable
    if not inspect.isgeneratorfunction(config.function) and not isinstance(
        results, tuple
    ):
        results = (results,)
    for result, write_proxy in zip(results, config.writes_map.values()):
        out_chunk_key = key_to_slices(
            out_coords_tuple, write_proxy.array, write_proxy.chunks
        )
        if isinstance(result, dict):  # group of arrays with named fields
            for k, v in result.items():
                v = backend_array_to_numpy_array(v)
                write_proxy.open().set_basic_selection(out_chunk_key, v, fields=k)
        else:
            result = backend_array_to_numpy_array(result)
            write_proxy.open()[out_chunk_key] = result

    if config.return_writes_stores:
        return [write_proxy.open().store for write_proxy in config.writes_map.values()]
    return None


def get_results_in_different_scope(out_coords: List[int], *, config: BlockwiseSpec):
    # wrap function call in a function so that args go out of scope (and free memory) as soon as results are returned

    # lithops needs params to be lists not tuples, so convert back
    out_coords_tuple = tuple(out_coords)

    # get array chunks for input keys, preserving any nested list structure
    get_chunk_config = partial(get_chunk, config=config)
    out_key = ("out",) + out_coords_tuple  # array name is ignored by key_function
    name_chunk_inds = list(config.key_function(out_key))
    args = map_nested(get_chunk_config, name_chunk_inds)

    return config.function(*args)


def key_to_slices(
    key: Tuple[int, ...], arr: T_ZarrArray, chunks: Optional[T_Chunks] = None
) -> Tuple[slice, ...]:
    """Convert a chunk index key to a tuple of slices"""
    chunks = normalize_chunks(chunks or arr.chunks, shape=arr.shape, dtype=arr.dtype)
    return get_item(chunks, key)


def get_chunk(in_key, config):
    """Read a chunk from the named array"""
    name = in_key[0]
    in_coords = in_key[1:]
    arr = config.reads_map[name].open()
    selection = key_to_slices(in_coords, arr)
    arg = arr[selection]
    arg = numpy_array_to_backend_array(arg)
    return arg


def blockwise(
    func: Callable[..., Any],
    out_ind: Sequence[Union[str, int]],
    *args: Any,
    allowed_mem: int,
    reserved_mem: int,
    target_store: T_Store,
    target_name: str,
    target_path: Optional[str] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    compressor: Union[dict, str, None] = "default",
    shape: T_Shape,
    dtype: T_DType,
    chunks: T_Chunks,
    new_axes: Optional[Dict[int, int]] = None,
    in_names: Optional[List[str]] = None,
    extra_projected_mem: int = 0,
    buffer_copies: Optional[BufferCopies] = None,
    extra_func_kwargs: Optional[Dict[str, Any]] = None,
    fusable_with_predecessors: bool = True,
    fusable_with_successors: bool = True,
    num_input_blocks: Optional[Tuple[int, ...]] = None,
    iterable_input_blocks: Optional[Tuple[bool, ...]] = None,
    **kwargs,
) -> PrimitiveOperation:
    """Apply a function to multiple blocks from multiple inputs, expressed using concise indexing rules.

    Unlike ```general_blockwise``, an index notation is used to specify the block mapping,
    like in Dask Array.

    Parameters
    ----------
    func : callable
        Function to apply to individual tuples of blocks
    out_ind : iterable
        Block pattern of the output, something like 'ijk' or (1, 2, 3)
    *args : sequence of Array, index pairs
        Sequence like (x, 'ij', y, 'jk', z, 'i')
    allowed_mem : int
        The memory available to a worker for running a task, in bytes. Includes ``reserved_mem``.
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes
    target_store : string or zarr.Array
        Path to output Zarr store, or Zarr array
    shape : tuple
        The shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        The chunks of the output array.
    new_axes : dict
        New indexes and their dimension lengths
    extra_projected_mem : int
        Extra memory projected to be needed (in bytes) in addition to the memory used reading
        the input arrays and writing the output.
    buffer_copies: BufferCopies
        The the number of buffer copies incurred for array storage operations.
    extra_func_kwargs : dict
        Extra keyword arguments to pass to function that can't be passed as regular keyword arguments
        since they clash with other blockwise arguments (such as dtype).
    **kwargs : dict
        Extra keyword arguments to pass to function

    Returns
    -------
    CubedPipeline to run the operation
    """

    arrays: Sequence[T_ZarrArray] = args[::2]
    array_names = in_names or [f"in_{i}" for i in range(len(arrays))]

    inds: Sequence[Union[str, int]] = args[1::2]

    numblocks: Dict[str, Tuple[int, ...]] = {}
    for name, array in zip(array_names, arrays, strict=True):
        input_chunks = normalize_chunks(
            array.chunks, shape=array.shape, dtype=array.dtype
        )
        numblocks[name] = tuple(map(len, input_chunks))

    argindsstr: List[Any] = []
    for name, ind in zip(array_names, inds, strict=True):
        argindsstr.extend((name, ind))

    key_function = make_blockwise_key_function_flattened(
        func,
        target_name,
        out_ind,
        *argindsstr,
        numblocks=numblocks,
        new_axes=new_axes,
    )

    return general_blockwise(
        func,
        key_function,
        *arrays,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        target_stores=[target_store],
        target_names=[target_name],
        target_paths=[target_path] if target_path is not None else None,
        storage_options=storage_options,
        compressor=compressor,
        shapes=[shape],
        dtypes=[dtype],
        chunkss=[chunks],
        in_names=in_names,
        extra_projected_mem=extra_projected_mem,
        buffer_copies=buffer_copies,
        extra_func_kwargs=extra_func_kwargs,
        fusable_with_predecessors=fusable_with_predecessors,
        fusable_with_successors=fusable_with_successors,
        num_input_blocks=num_input_blocks,
        iterable_input_blocks=iterable_input_blocks,
        **kwargs,
    )


def general_blockwise(
    func: Callable[..., Any],
    key_function: Callable[..., Any],
    *arrays: Any,
    allowed_mem: int,
    reserved_mem: int,
    target_stores: List[T_Store],
    target_names: List[str],
    target_paths: Optional[List[str]] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    compressor: Union[dict, str, None] = "default",
    shapes: List[T_Shape],
    dtypes: List[T_DType],
    chunkss: List[T_Chunks],
    in_names: Optional[List[str]] = None,
    extra_projected_mem: int = 0,
    buffer_copies: Optional[BufferCopies] = None,
    extra_func_kwargs: Optional[Dict[str, Any]] = None,
    fusable_with_predecessors: bool = True,
    fusable_with_successors: bool = True,
    function_nargs: Optional[int] = None,
    num_input_blocks: Optional[Tuple[int, ...]] = None,
    iterable_input_blocks: Optional[Tuple[bool, ...]] = None,
    target_chunks_: Optional[T_RegularChunks] = None,
    return_writes_stores: bool = False,
    output_blocks: Optional[Iterator[List[int]]] = None,
    **kwargs,
) -> PrimitiveOperation:
    """A more general form of ``blockwise`` that uses a function to specify the block
    mapping, rather than an index notation, and which supports multiple outputs.

    For multiple outputs, all output arrays must have matching numblocks.

    Parameters
    ----------
    func : callable
        Function to apply to individual tuples of blocks
    key_function : callable
        A function that maps an output chunk key to one or more input chunk keys.
    *arrays : sequence of Array
        The input arrays.
    allowed_mem : int
        The memory available to a worker for running a task, in bytes. Includes ``reserved_mem``.
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes
    target_store : string or zarr.Array
        Path to output Zarr store, or Zarr array
    shape : tuple
        The shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        The chunks of the output array.
    extra_projected_mem : int
        Extra memory projected to be needed (in bytes) in addition to the memory used reading
        the input arrays and writing the output.
    buffer_copies: BufferCopies
        The the number of buffer copies incurred for array storage operations.
    extra_func_kwargs : dict
        Extra keyword arguments to pass to function that can't be passed as regular keyword arguments
        since they clash with other blockwise arguments (such as dtype).
    **kwargs : dict
        Extra keyword arguments to pass to function

    Returns
    -------
    PrimitiveOperation to run the operation
    """
    array_names = in_names or [f"in_{i}" for i in range(len(arrays))]
    array_map = {name: array for name, array in zip(array_names, arrays, strict=True)}

    func_kwargs = extra_func_kwargs or {}
    func_with_kwargs = partial(func, **{**kwargs, **func_kwargs})
    function_nargs = function_nargs or len(arrays)
    num_input_blocks = num_input_blocks or (1,) * len(arrays)
    iterable_input_blocks = iterable_input_blocks or (False,) * len(arrays)
    read_proxies = {
        name: CubedArrayProxy(array, array.chunks) for name, array in array_map.items()
    }

    write_proxies = {}
    output_chunk_memory = 0
    target_arrays = []

    numblocks0 = None
    for i, target_store in enumerate(target_stores):
        chunks_normal = normalize_chunks(chunkss[i], shape=shapes[i], dtype=dtypes[i])
        chunksize = to_chunksize(chunks_normal)
        if numblocks0 is None:
            numblocks0 = compute_numblocks(chunks_normal)
        else:
            numblocks = compute_numblocks(chunks_normal)
            if numblocks != numblocks0:
                raise ValueError(
                    f"All outputs must have matching number of blocks in each dimension. Chunks specified: {chunkss}"
                )
        ta: Union[zarr.Array, LazyZarrArray]
        if is_backend_storage_array(target_store) or isinstance(
            target_store, LazyZarrArray
        ):
            ta = target_store
        else:
            ta = lazy_zarr_array(
                target_store,
                shapes[i],
                dtype=dtypes[i],
                chunks=target_chunks_ or chunksize,
                path=target_names[i],  # use array name for path within store
                storage_options=storage_options,
                compressor=compressor,
            )
        target_arrays.append(ta)

        write_proxies[target_names[i]] = CubedArrayProxy(ta, chunksize)

        # only one output chunk is read into memory at a time, so we find the largest
        output_chunk_memory = max(
            output_chunk_memory, array_memory(dtypes[i], chunksize)
        )

    # the number of blocks written to each target array is currently the same
    nb = math.prod(chunksize) // math.prod(target_chunks_ or chunksize)
    num_output_blocks = (nb,) * len(target_arrays)

    spec = BlockwiseSpec(
        key_function,
        func_with_kwargs,
        function_nargs,
        num_input_blocks,
        num_output_blocks,
        iterable_input_blocks,
        read_proxies,
        write_proxies,
        return_writes_stores,
    )

    buffer_copies = buffer_copies or BufferCopies(read=1, write=1)
    projected_mem = calculate_projected_mem(
        reserved_mem=reserved_mem,
        inputs=[array_memory(array.dtype, array.chunks) for array in arrays],
        operation=extra_projected_mem,
        output=output_chunk_memory,
        buffer_copies=buffer_copies,
    )

    if projected_mem > allowed_mem:
        raise ValueError(
            f"Projected blockwise memory ({projected_mem}) exceeds allowed_mem ({allowed_mem}), including reserved_mem ({reserved_mem})"
        )

    # this must be an iterator of lists, not of tuples, otherwise lithops breaks
    if output_blocks is None:
        output_blocks = map(
            list, itertools.product(*[range(len(c)) for c in chunks_normal])
        )
    num_tasks = math.prod(len(c) for c in chunks_normal)

    pipeline = CubedPipeline(
        apply_blockwise,
        gensym("apply_blockwise"),
        output_blocks,
        spec,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=array_names,
        target_array=target_arrays[0] if len(target_arrays) == 1 else target_arrays,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable_with_predecessors=fusable_with_predecessors,
        fusable_with_successors=fusable_with_successors,
        write_chunks=chunksize,
    )


# Code for fusing blockwise operations


def is_fuse_candidate(primitive_op: PrimitiveOperation) -> bool:
    """
    Return True if a primitive operation is a candidate for blockwise fusion.
    """
    return primitive_op.pipeline.function == apply_blockwise and (
        primitive_op.fusable_with_predecessors or primitive_op.fusable_with_successors
    )


def can_fuse_primitive_ops(
    primitive_op1: PrimitiveOperation, primitive_op2: PrimitiveOperation
) -> bool:
    if is_fuse_candidate(primitive_op1) and is_fuse_candidate(primitive_op2):
        return primitive_op1.num_tasks == primitive_op2.num_tasks
    return False


def can_fuse_multiple_primitive_ops(
    name: str,
    primitive_op: PrimitiveOperation,
    predecessor_primitive_ops: List[PrimitiveOperation],
    *,
    max_total_num_input_blocks: Optional[int] = None,
) -> bool:
    if is_fuse_candidate(primitive_op) and all(
        p is None or is_fuse_candidate(p) for p in predecessor_primitive_ops
    ):
        # If the peak projected memory for running all the predecessor ops in
        # order is larger than allowed_mem then we can't fuse.
        peak_projected = peak_projected_mem(predecessor_primitive_ops)
        if peak_projected > primitive_op.allowed_mem:
            logger.debug(
                "can't fuse %s since peak projected memory for predecessor ops (%s) is greater than allowed (%s)",
                name,
                peak_projected,
                primitive_op.allowed_mem,
            )
            return False
        if max_total_num_input_blocks is None:
            # If max total input blocks not specified, then only fuse if num
            # tasks of predecessor ops match.
            ret = all(
                primitive_op.num_tasks == p.num_tasks
                for p in predecessor_primitive_ops
                if p is not None
            )
            if ret:
                logger.debug(
                    "can fuse %s since num tasks of predecessor ops match", name
                )
            else:
                logger.debug(
                    "can't fuse %s since num tasks of predecessor ops do not match",
                    name,
                )
            return ret
        else:
            num_input_blocks = primitive_op.pipeline.config.num_input_blocks
            total_num_input_blocks = 0
            for ni, p in zip(num_input_blocks, predecessor_primitive_ops, strict=True):
                if p is None:
                    continue
                for nj in p.pipeline.config.num_input_blocks:
                    total_num_input_blocks += ni * nj
            ret = total_num_input_blocks <= max_total_num_input_blocks
            if ret:
                logger.debug(
                    "can fuse %s since total number of input blocks (%s) does not exceed max (%s)",
                    name,
                    total_num_input_blocks,
                    max_total_num_input_blocks,
                )
            else:
                logger.debug(
                    "can't fuse %s since total number of input blocks (%s) exceeds max (%s)",
                    name,
                    total_num_input_blocks,
                    max_total_num_input_blocks,
                )
            return ret
    logger.debug(
        "can't fuse %s since primitive op and predecessors are not all candidates", name
    )
    return False


def peak_projected_mem(primitive_ops):
    """Calculate the peak projected memory for running a series of primitive ops
    and retaining their return values in memory."""
    memory_modeller = MemoryModeller()
    for p in primitive_ops:
        if p is None:
            continue
        memory_modeller.allocate(p.projected_mem)
        chunkmem = chunk_memory(p.target_array)
        memory_modeller.free(p.projected_mem - chunkmem)
    return memory_modeller.peak_mem


def fuse(
    primitive_op1: PrimitiveOperation, primitive_op2: PrimitiveOperation
) -> PrimitiveOperation:
    """
    Fuse two blockwise operations into a single operation, avoiding writing to (or reading from) the target of the first operation.
    """

    assert primitive_op1.num_tasks == primitive_op2.num_tasks

    pipeline1 = primitive_op1.pipeline
    pipeline2 = primitive_op2.pipeline

    mappable = pipeline2.mappable

    def fused_key_func(out_key):
        return pipeline1.config.key_function(*pipeline2.config.key_function(out_key))

    def fused_func(*args):
        return pipeline2.config.function(pipeline1.config.function(*args))

    function_nargs = pipeline1.config.function_nargs
    read_proxies = pipeline1.config.reads_map
    write_proxies = pipeline2.config.writes_map
    return_writes_stores = pipeline2.config.return_writes_stores
    num_input_blocks = tuple(
        n * pipeline2.config.num_input_blocks[0]
        for n in pipeline1.config.num_input_blocks
    )
    num_output_blocks = pipeline2.config.num_output_blocks
    iterable_input_blocks = pipeline1.config.iterable_input_blocks
    spec = BlockwiseSpec(
        fused_key_func,
        fused_func,
        function_nargs,
        num_input_blocks,
        num_output_blocks,
        iterable_input_blocks,
        read_proxies,
        write_proxies,
        return_writes_stores,
    )

    source_array_names = primitive_op1.source_array_names
    target_array = primitive_op2.target_array
    projected_mem = max(primitive_op1.projected_mem, primitive_op2.projected_mem)
    allowed_mem = primitive_op2.allowed_mem
    reserved_mem = primitive_op2.reserved_mem
    num_tasks = primitive_op2.num_tasks

    pipeline = CubedPipeline(
        apply_blockwise,
        gensym("fused_apply_blockwise"),
        mappable,
        spec,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=source_array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable_with_predecessors=True,
    )


def fuse_multiple(
    primitive_op: PrimitiveOperation, *predecessor_primitive_ops: PrimitiveOperation
) -> PrimitiveOperation:
    """
    Fuse a blockwise operation and its predecessors into a single operation, avoiding writing to (or reading from) the targets of the predecessor operations.
    """

    bw_spec = primitive_op.pipeline.config

    null_blockwise_spec = BlockwiseSpec(
        key_function=lambda x: (x,),
        function=lambda x: x,
        function_nargs=1,
        num_input_blocks=(1,),
        num_output_blocks=(1,),
        iterable_input_blocks=(False,),
        reads_map={},
        writes_map={},
    )
    predecessor_bw_specs = [
        primitive_op.pipeline.config
        if primitive_op is not None
        else null_blockwise_spec
        for primitive_op in predecessor_primitive_ops
    ]

    spec = fuse_blockwise_specs(bw_spec, *predecessor_bw_specs)

    source_array_names = []
    for i, p in enumerate(predecessor_primitive_ops):
        if p is None:
            source_array_names.append(primitive_op.source_array_names[i])
        else:
            source_array_names.extend(p.source_array_names)
    target_array = primitive_op.target_array
    projected_mem = max(
        primitive_op.projected_mem,
        peak_projected_mem(p for p in predecessor_primitive_ops if p is not None),
    )
    allowed_mem = primitive_op.allowed_mem
    reserved_mem = primitive_op.reserved_mem
    num_tasks = primitive_op.num_tasks

    fused_pipeline = CubedPipeline(
        apply_blockwise,
        gensym("fused_apply_blockwise"),
        primitive_op.pipeline.mappable,
        spec,
    )
    return PrimitiveOperation(
        pipeline=fused_pipeline,
        source_array_names=source_array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable_with_predecessors=True,
    )


def fuse_blockwise_specs(
    bw_spec: BlockwiseSpec, *predecessor_bw_specs: BlockwiseSpec
) -> BlockwiseSpec:
    """
    Fuse a blockwise spec and its predecessors into a single spec.
    """

    predecessor_funcs_nargs = [bws.function_nargs for bws in predecessor_bw_specs]

    fused_key_func = make_fused_key_function(
        bw_spec.key_function,
        [bws.key_function for bws in predecessor_bw_specs],
        predecessor_funcs_nargs,
    )

    fused_func = make_fused_function(
        bw_spec.function,
        [bws.function for bws in predecessor_bw_specs],
        bw_spec.iterable_input_blocks,
    )

    fused_function_nargs = bw_spec.function_nargs
    num_input_blocks = bw_spec.num_input_blocks
    predecessor_num_blocks = [bws.num_input_blocks for bws in predecessor_bw_specs]
    # note that the length of fused_num_input_blocks is the same as the number of input arrays
    # but may be different to fused_function_nargs since the fused function groups args
    fused_num_input_blocks = tuple(
        itertools.chain(
            *(
                tuple(n * m for m in nb)
                for n, nb in zip(num_input_blocks, predecessor_num_blocks, strict=True)
            )
        )
    )
    fused_num_output_blocks = bw_spec.num_output_blocks
    predecessor_iterable_input_blocks = [
        bws.iterable_input_blocks for bws in predecessor_bw_specs
    ]
    fused_iterable_input_blocks = tuple(
        itertools.chain(*predecessor_iterable_input_blocks)
    )

    read_proxies = dict(bw_spec.reads_map)
    for bws in predecessor_bw_specs:
        read_proxies.update(bws.reads_map)
    write_proxies = bw_spec.writes_map
    return_writes_stores = bw_spec.return_writes_stores
    return BlockwiseSpec(
        fused_key_func,
        fused_func,
        fused_function_nargs,
        fused_num_input_blocks,
        fused_num_output_blocks,
        fused_iterable_input_blocks,
        read_proxies,
        write_proxies,
        return_writes_stores,
    )


def apply_blockwise_key_func(key_function, arg):
    if isinstance(arg, tuple):
        return key_function(arg)
    else:
        # more than one input block is being read from arg
        assert isinstance(arg, (list, Iterator))
        if isinstance(arg, list):
            return tuple(
                list(item) for item in zip(*(key_function(a) for a in arg), strict=True)
            )
        else:
            # Return iterators to avoid materializing all array blocks at
            # once.
            return tuple(
                iter(list(item))
                for item in zip(*(key_function(a) for a in arg), strict=True)
            )


def apply_blockwise_func(func, is_iterable, *args):
    if is_iterable is False:
        ret = func(*args)
    else:
        # More than one input block is being read from this group of args to primitive op.
        # Note that it is important that a list is not returned to avoid materializing all
        # array blocks at once.
        ret = map(lambda item: func(*item), zip(*args, strict=True))
    return ret


def make_fused_key_function(
    key_function, predecessor_key_functions, predecessor_funcs_nargs
):
    def fused_key_func(out_key):
        args = key_function(out_key)
        # split all args to the fused function into groups, one for each predecessor function
        func_args = tuple(
            item
            for pkf, a in zip(predecessor_key_functions, args, strict=True)
            for item in apply_blockwise_key_func(pkf, a)
        )
        return split_into(func_args, predecessor_funcs_nargs)

    return fused_key_func


def make_fused_function(function, predecessor_functions, iterable_input_blocks):
    def fused_func_single(*args):
        # args are grouped appropriately so they can be called by each predecessor function
        func_args = [
            apply_blockwise_func(pf, iterable_input_blocks[i], *a)
            for i, (pf, a) in enumerate(zip(predecessor_functions, args, strict=True))
        ]
        return function(*func_args)

    # multiple outputs
    def fused_func_generator(*args):
        # args are grouped appropriately so they can be called by each predecessor function
        func_args = [
            apply_blockwise_func(pf, iterable_input_blocks[i], *a)
            for i, (pf, a) in enumerate(zip(predecessor_functions, args, strict=True))
        ]
        yield from function(*func_args)

    return (
        fused_func_generator
        if inspect.isgeneratorfunction(function)
        else fused_func_single
    )


# blockwise key functions


def make_blockwise_key_function(
    func: Callable[..., Any],
    output: str,
    out_indices: Sequence[Union[str, int]],
    *arrind_pairs: Any,
    numblocks: Dict[str, Tuple[int, ...]],
    new_axes: Optional[Dict[int, int]] = None,
) -> Callable[[List[int]], Any]:
    """Make a function that is the equivalent of make_blockwise_graph."""

    new_axes = new_axes or {}
    argpairs = list(toolz.partition(2, arrind_pairs))

    # Dictionary mapping {i: 3, j: 4, ...} for i, j, ... the dimensions
    dims = _make_dims(argpairs, numblocks, new_axes)

    # Generate the abstract "plan" before constructing
    # the actual graph
    (coord_maps, concat_axes, dummies) = _get_coord_mapping(
        dims,
        output,
        out_indices,
        numblocks,
        argpairs,
        False,
    )

    for axes, (arg, _) in zip(concat_axes, argpairs, strict=True):
        for ax in axes:
            if numblocks[arg][ax] > 1:
                raise ValueError(
                    f"Cannot have multiple chunks in dropped axis {ax}. "
                    "To fix, use a reduction after calling map_blocks "
                    "without specifying drop_axis, or rechunk first."
                )

    def key_function(out_key):
        out_coords = out_key[1:]

        # from Dask make_blockwise_graph
        deps = set()
        coords = out_coords + dummies
        args = []
        for cmap, axes, (arg, ind) in zip(
            coord_maps, concat_axes, argpairs, strict=True
        ):
            if ind is None:
                args.append(arg)
            else:
                arg_coords = tuple(coords[c] for c in cmap)
                if axes:
                    tups = lol_product((arg,), arg_coords)
                    deps.update(flatten(tups))
                else:
                    tups = (arg,) + arg_coords
                    deps.add(tups)
                args.append(tups)

        args.insert(0, func)
        val = tuple(args)
        # end from make_blockwise_graph

        return val

    return key_function


def make_blockwise_key_function_flattened(
    func: Callable[..., Any],
    output: str,
    out_indices: Sequence[Union[str, int]],
    *arrind_pairs: Any,
    numblocks: Dict[str, Tuple[int, ...]],
    new_axes: Optional[Dict[int, int]] = None,
) -> Callable[[List[int]], Any]:
    # TODO: make this a part of make_blockwise_key_function?
    key_function = make_blockwise_key_function(
        func, output, out_indices, *arrind_pairs, numblocks=numblocks, new_axes=new_axes
    )

    def blockwise_fn_flattened(out_key):
        in_keys = key_function(out_key)[1:]  # drop function in position 0
        # flatten (nested) lists indicating contraction
        if isinstance(in_keys[0], list):
            in_keys = list(flatten(in_keys))
        return in_keys

    return blockwise_fn_flattened



import re

# Modified version of `numpy.lib.function_base._parse_gufunc_signature`
# Modifications:
#   - Allow for zero input arguments
# See https://docs.scipy.org/doc/numpy/reference/c-api/generalized-ufuncs.html
_DIMENSION_NAME = r"\w+"
_CORE_DIMENSION_LIST = "(?:{0:}(?:,{0:})*,?)?".format(_DIMENSION_NAME)
_ARGUMENT = rf"\({_CORE_DIMENSION_LIST}\)"
_INPUT_ARGUMENTS = "(?:{0:}(?:,{0:})*,?)?".format(_ARGUMENT)
_OUTPUT_ARGUMENTS = "{0:}(?:,{0:})*".format(
    _ARGUMENT
)  # Use `'{0:}(?:,{0:})*,?'` if gufunc-
# signature should be allowed for length 1 tuple returns
_SIGNATURE = f"^{_INPUT_ARGUMENTS}->{_OUTPUT_ARGUMENTS}$"


def _parse_gufunc_signature(signature):
    """
    Parse string signatures for a generalized universal function.

    Arguments
    ---------
    signature : string
        Generalized universal function signature, e.g., ``(m,n),(n,p)->(m,p)``
        for ``np.matmul``.

    Returns
    -------
    Tuple of input and output core dimensions parsed from the signature, each
    of the form List[Tuple[str, ...]], except for one output. For one output
    core dimension is not a list, but of the form Tuple[str, ...]
    """
    signature = re.sub(r"\s+", "", signature)
    if not re.match(_SIGNATURE, signature):
        raise ValueError(f"Not a valid gufunc signature: {signature}")
    in_txt, out_txt = signature.split("->")
    ins = [
        tuple(re.findall(_DIMENSION_NAME, arg)) for arg in re.findall(_ARGUMENT, in_txt)
    ]
    outs = [
        tuple(re.findall(_DIMENSION_NAME, arg))
        for arg in re.findall(_ARGUMENT, out_txt)
    ]
    outs = outs[0] if ((len(outs) == 1) and (out_txt[-1] != ",")) else outs
    return ins, outs



from typing import TYPE_CHECKING

from cubed.array_api.manipulation_functions import broadcast_to, expand_dims
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import map_blocks, map_selection, reduction
from cubed.utils import get_item, normalize_chunks

if TYPE_CHECKING:
    from cubed.array_api.array_object import Array


def groupby_reduction(
    x: "Array",
    by: "Array",
    func,
    combine_func=None,
    aggregate_func=None,
    axis=None,
    intermediate_dtype=None,
    dtype=None,
    keepdims=False,
    split_every=None,
    num_groups=None,
    extra_func_kwargs=None,
) -> "Array":
    """A reduction operation that performs groupby aggregations.

    Parameters
    ----------
    x: Array
        Array being grouped along one axis.
    by: Array
        Array of non-negative integers to be used as labels with which to group
        the values in ``x`` along the reduction axis. Must be a 1D array.
    func: callable
        Function to apply to each chunk of data before reduction.
    combine_func: callable
        Function which may be applied recursively to intermediate chunks of
        data. The number of chunks that are combined in each round is
        determined by the ``split_every`` parameter. The output of the
        function is a chunk with size ``num_groups`` along the reduction axis.
    aggregate_func: callable, optional
        Function to apply to each of the final chunks to produce the final output.
    axis: int or sequence of ints, optional
        Axis to aggregate along. Only supports a single axis.
    intermediate_dtype: dtype
        Data type of intermediate output.
    dtype: dtype
        Data type of output.
    keepdims: boolean, optional
        Whether the reduction function should preserve the reduced axes,
        or remove them.
    split_every: int >= 2 or dict(axis: int), optional
        The number of chunks to combine in one round along each axis in the
        recursive aggregation.
    num_groups: int
        The number of groups in the grouping array ``by``.
    extra_func_kwargs: dict, optional
        Extra keyword arguments to pass to ``func`` and ``combine_func``.
    """

    if isinstance(axis, tuple):
        if len(axis) != 1:
            raise ValueError(
                f"Only a single axis is supported for groupby_reduction: {axis}"
            )
        axis = axis[0]

    # make sure 'by' has corresponding blocks to 'x'
    for ax in range(x.ndim):
        if ax != axis:
            by = expand_dims(by, axis=ax)
    by_chunks = tuple(
        c if i == axis else (1,) * x.numblocks[i] for i, c in enumerate(by.chunks)
    )
    by_shape = tuple(map(sum, by_chunks))
    by = broadcast_to(by, by_shape, chunks=by_chunks)

    # wrapper to squeeze 'by' to undo effect of broadcast, so it looks same
    # to user supplied func
    def _group_reduction_func_wrapper(func):
        def wrapper(a, by, **kwargs):
            return func(a, nxp.squeeze(by), **kwargs)

        return wrapper

    # initial map does group reduction on each block
    chunks = tuple(
        (num_groups,) * len(c) if i == axis else c for i, c in enumerate(x.chunks)
    )
    out = map_blocks(
        _group_reduction_func_wrapper(func),
        x,
        by,
        dtype=intermediate_dtype,
        chunks=chunks,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        num_groups=num_groups,
    )

    # add a dummy dimension to reduce over
    dummy_axis = -1
    out = expand_dims(out, axis=dummy_axis)

    # then reduce across blocks
    return reduction(
        out,
        func=None,
        combine_func=combine_func,
        aggregate_func=aggregate_func,
        axis=(dummy_axis, axis),  # dummy and group axis
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
        combine_sizes={axis: num_groups},  # group axis doesn't have size 1
        extra_func_kwargs=dict(dtype=intermediate_dtype, dummy_axis=dummy_axis),
    )


def groupby_blockwise(
    x: "Array",
    by,
    func,
    axis=None,
    dtype=None,
    num_groups=None,
    **kwargs,
):
    """A blockwise operation that performs groupby aggregations.

    Parameters
    ----------
    x: Array
        Array being grouped along one axis.
    by: nxp.array
        Array of non-negative integers to be used as labels with which to group
        the values in ``x`` along the reduction axis. Must be a 1D array.
    func: callable
        Function to apply to each chunk of data. The output of the
        function is a chunk with size corresponding to the number of groups in the
        input chunk along the reduction axis.
    axis: int or sequence of ints, optional
        Axis to aggregate along. Only supports a single axis.
    dtype: dtype
        Data type of output.
    num_groups: int
        The number of groups in the grouping array ``by``.
    """

    if by.ndim != 1:
        raise ValueError(f"Array `by` must be 1D, but has {by.ndim} dimensions.")

    if isinstance(axis, tuple):
        if len(axis) != 1:
            raise ValueError(
                f"Only a single axis is supported for groupby_reduction: {axis}"
            )
        axis = axis[0]

    newchunks, groups_per_chunk = _get_chunks_for_groups(
        x.numblocks[axis],
        by,
        num_groups=num_groups,
    )

    # calculate the chunking used to read the input array 'x'
    read_chunks = tuple(newchunks if i == axis else c for i, c in enumerate(x.chunks))

    # 'by' is not a cubed array, but we still read it in chunks
    by_read_chunks = (newchunks,)

    # find shape and chunks for the output
    shape = tuple(num_groups if i == axis else s for i, s in enumerate(x.shape))
    chunks = tuple(
        groups_per_chunk if i == axis else c for i, c in enumerate(x.chunksize)
    )
    target_chunks = normalize_chunks(chunks, shape, dtype=dtype)

    def selection_function(out_key):
        out_coords = out_key[1:]
        block_id = out_coords
        return get_item(read_chunks, block_id)

    # in general each selection overlaps 2 input blocks
    max_num_input_blocks = 2

    return map_selection(
        _process_blockwise_chunk,
        selection_function,
        x,
        shape=shape,
        dtype=dtype,
        chunks=target_chunks,
        max_num_input_blocks=max_num_input_blocks,
        axis=axis,
        by=by,
        blockwise_func=func,
        by_read_chunks=by_read_chunks,
        target_chunks=target_chunks,
        groups_per_chunk=groups_per_chunk,
        **kwargs,
    )


def _process_blockwise_chunk(
    a,
    axis=None,
    by=None,
    blockwise_func=None,
    by_read_chunks=None,
    target_chunks=None,
    groups_per_chunk=None,
    block_id=None,
    **kwargs,
):
    idx = block_id
    bi = idx[axis]

    by = by[get_item(by_read_chunks, (bi,))]

    start_group = bi * groups_per_chunk

    return blockwise_func(
        a,
        by,
        axis=axis,
        start_group=start_group,
        num_groups=target_chunks[axis][bi],
        **kwargs,
    )


def _get_chunks_for_groups(num_chunks, labels, num_groups):
    """Find new chunking so that there are an equal number of group labels per chunk."""

    # find the start indexes of each group
    start_indexes = nxp.searchsorted(labels, nxp.arange(num_groups))

    # find the number of groups per chunk
    groups_per_chunk = max(num_groups // num_chunks, 1)

    # each chunk has groups_per_chunk groups in it (except possibly last one)
    chunk_boundaries = start_indexes[::groups_per_chunk]

    # successive differences give the new chunk sizes (include end index for last chunk)
    newchunks = nxp.diff(chunk_boundaries, append=len(labels))

    return tuple(newchunks), groups_per_chunk



from typing import Any, Optional, Sequence

from networkx import MultiDiGraph
from pyspark.sql import SparkSession

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import handle_callbacks, handle_operation_start_callbacks
from cubed.spec import Spec


class SparkExecutor(DagExecutor):
    """An execution engine that uses Apache Spark."""

    # Minimum memory allowed for Spark (512MB)
    MIN_MEMORY_MiB = 512

    def __init__(self, **kwargs):
        self._callbacks = None
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "spark"

    def _parse_memory_setting(self, mem_bytes: int):
        """
        Convert memory value to a valid Spark memory setting string.
        Ensures the value is at least MIN_MEMORY_MiB.

        Args:
            mem_bytes: Memory value in bytes

        Returns:
            String memory setting suitable for Spark config
            with a size unit suffix ("k", "m", "g" or "t")
            (e.g. 512m, 2g).
        """
        # Try to convert to bytes if it's a number
        try:
            # Convert to MiB and ensure minimum threshold
            mb_value = max(self.MIN_MEMORY_MiB, mem_bytes // (1024 * 1024))
            return f"{mb_value}m"
        except (ValueError, TypeError):
            # If conversion fails, return default
            return f"{self.MIN_MEMORY_MiB}m"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs: Any,
    ):
        # Store callbacks for later use during computation
        self._callbacks = callbacks

        # Configure Spark memory settings from Spec if provided
        spark_builder = SparkSession.builder
        if spec is not None and hasattr(spec, "allowed_mem") and spec.allowed_mem:
            mem_setting = self._parse_memory_setting(spec.allowed_mem)
            spark_builder = spark_builder.config("spark.executor.memory", mem_setting)
            spark_builder = spark_builder.config("spark.driver.memory", mem_setting)
            spark_builder = spark_builder.config("spark.speculation", "true")

        # Create a Spark session
        spark = spark_builder.getOrCreate()

        for name, node in visit_nodes(dag):
            handle_operation_start_callbacks(callbacks, name)
            pipeline = node["pipeline"]
            # Create an RDD from pipeline.mappable.
            rdd = spark.sparkContext.parallelize(pipeline.mappable)
            # Define the transformation; note that this is lazy.
            lazy_rdd = rdd.map(
                lambda x, pipeline=pipeline: pipeline.function(
                    x, config=pipeline.config
                )
            )
            results = lazy_rdd.collect()  # <-- Trigger computation immediately
            if callbacks is not None:
                for result in results:
                    handle_callbacks(callbacks, result, {"name": name})



import cubed.array_api as xp

info = xp.__array_namespace_info__()


def test_capabilities():
    capabilities = info.capabilities()
    assert capabilities["boolean indexing"] is False
    assert capabilities["data-dependent shapes"] is False


def test_default_device():
    assert (
        info.default_device() is None or info.default_device() == xp.asarray(0).device
    )


def test_default_dtypes():
    dtypes = info.default_dtypes()
    assert dtypes["real floating"] == xp.asarray(0.0).dtype
    assert dtypes["complex floating"] == xp.asarray(0.0j).dtype
    assert dtypes["integral"] == xp.asarray(0).dtype
    assert dtypes["indexing"] == xp.argmax(xp.zeros(10)).dtype


def test_devices():
    assert len(info.devices()) > 0


def test_dtypes():
    assert len(info.dtypes()) > 0



from typing import Any, Dict

import networkx as nx


def skip_node(name, dag, nodes: Dict[str, Any]) -> bool:
    """
    Return True if the array for a node doesn't have a pipeline to compute it,
    or if it is marked as already computed.
    """
    pipeline = nodes[name].get("pipeline", None)
    if pipeline is None:
        return True

    return nodes[name].get("computed", False)


def visit_nodes(dag):
    """Return a generator that visits the nodes in the DAG in topological order."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for name in list(nx.topological_sort(dag)):
        if skip_node(name, dag, nodes):
            continue
        yield name, nodes[name]


def visit_node_generations(dag):
    """Return a generator that visits the nodes in the DAG in groups of topological generations."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for names in nx.topological_generations(dag):
        gen = [(name, nodes[name]) for name in names if not skip_node(name, dag, nodes)]
        if len(gen) > 0:
            yield gen



import numpy as np
import pytest

import cubed.array_api as xp
from cubed._testing import assert_array_equal
from cubed.core.ops import smallest_blockdim, unify_chunks
from cubed.tests.utils import TaskCounter


def test_smallest_blockdim():
    assert smallest_blockdim([]) == ()
    assert smallest_blockdim([(5,), (5,)]) == (5,)
    assert smallest_blockdim([(5,), (3, 2)]) == (3, 2)
    assert smallest_blockdim([(5, 5), (3, 3, 3, 1)]) == (3, 3, 3, 1)
    assert smallest_blockdim([(2, 1), (2, 1)]) == (2, 1)
    assert smallest_blockdim([(2, 2, 1), (3, 2), (2, 2, 1)]) == (2, 2, 1)

    with pytest.raises(ValueError, match="Chunks do not add up to same value"):
        smallest_blockdim([(2, 1), (2, 2)])


@pytest.mark.parametrize(
    ("chunks_a", "chunks_b", "expected_chunksize"),
    [
        ((2,), (4,), (2,)),
        ((4,), (2,), (2,)),
        ((6,), (10,), (6,)),
        ((10,), (10,), (10,)),
        ((5,), (10,), (5,)),
        ((3,), (5,), (3,)),
        ((5,), (3,), (3,)),
    ],
)
def test_unify_chunks_elemwise(chunks_a, chunks_b, expected_chunksize):
    a = xp.ones((10,), chunks=chunks_a)
    b = xp.ones((10,), chunks=chunks_b)

    _, arrays = unify_chunks(a, "i", b, "i")
    for arr in arrays:
        assert arr.chunksize == expected_chunksize

    c = xp.add(a, b)
    assert_array_equal(c.compute(), np.ones((10,)) + np.ones((10,)))


@pytest.mark.parametrize(
    ("chunks_a", "chunks_b", "expected_chunksize"),
    [
        ((2, 2), (4, 4), (2, 2)),
        ((2, 4), (4, 2), (2, 2)),
        ((4, 2), (2, 4), (2, 2)),
        ((3, 5), (5, 3), (3, 3)),
        ((3, 10), (10, 3), (3, 3)),
    ],
)
def test_unify_chunks_elemwise_2d(chunks_a, chunks_b, expected_chunksize):
    a = xp.ones((10, 10), chunks=chunks_a)
    b = xp.ones((10, 10), chunks=chunks_b)

    _, arrays = unify_chunks(a, "ij", b, "ij")
    for arr in arrays:
        assert arr.chunksize == expected_chunksize

    c = xp.add(a, b)
    assert_array_equal(c.compute(), np.ones((10, 10)) + np.ones((10, 10)))


@pytest.mark.parametrize(
    ("chunks_a", "chunks_b", "expected_chunksize"),
    [
        ((2, 2), (4, 4), (2, 2)),
        ((2, 4), (2, 4), (2, 2)),
        ((4, 2), (4, 2), (2, 2)),
        ((3, 5), (3, 5), (3, 3)),
        ((3, 10), (3, 10), (3, 3)),
    ],
)
def test_unify_chunks_blockwise_2d(chunks_a, chunks_b, expected_chunksize):
    a = xp.ones((10, 10), chunks=chunks_a)
    b = xp.ones((10, 10), chunks=chunks_b)

    _, arrays = unify_chunks(a, "ij", b, "ji")
    for arr in arrays:
        assert arr.chunksize == expected_chunksize

    c = xp.matmul(a, b)
    assert_array_equal(c.compute(), np.matmul(np.ones((10, 10)), np.ones((10, 10))))


def test_unify_chunks_broadcast_scalar():
    a = xp.ones((10,), chunks=(3,))
    b = a + 1
    assert_array_equal(b.compute(), np.ones((10,)) + 1)


def test_unify_chunks_broadcast_2d():
    a = xp.ones((10, 10), chunks=(3, 3))
    b = xp.ones((10,), chunks=(5,))
    c = xp.add(a, b)

    # the following checks that b is rechunked *before* broadcasting, to avoid materializing the full (broadcasted) array
    task_counter = TaskCounter()
    res = c.compute(callbacks=[task_counter])
    num_created_arrays = 2  # b rechunked, c
    # 1 task for rechunk of b, 16 for addition operation
    assert task_counter.value == num_created_arrays + 1 + 16

    assert_array_equal(res, np.ones((10, 10)) + np.ones((10,)))



from typing import Optional, Union

import zarr

from cubed.storage.backend import open_backend_array
from cubed.storage.types import ArrayMetadata
from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store


class LazyZarrArray(ArrayMetadata):
    """A Zarr array that may not have been written to storage yet.

    On creation, a normal Zarr array's metadata is immediately written to storage,
    and there is no way to avoid this, hence this class, which separates the creation
    of the object in memory and the creation of array metadata in storage.
    """

    def __init__(
        self,
        store: T_Store,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
        path: Optional[str] = None,
        **kwargs,
    ):
        """Create a Zarr array lazily in memory."""
        super().__init__(shape, dtype, chunks)
        self.store = store
        self.path = path
        self.kwargs = kwargs

    def create(self, mode: str = "w-") -> zarr.Array:
        """Create the Zarr array in storage.

        The Zarr array's metadata is initialized in the backing store, and any
        initial values are set on the array.

        Parameters
        ----------
        mode : str
            The mode to open the Zarr array with using ``zarr.open``.
            Default is 'w-', which means create, fail it already exists.
        """
        target = open_backend_array(
            self.store,
            mode=mode,
            shape=self.shape,
            dtype=self.dtype,
            chunks=self.chunks,
            path=self.path,
            **self.kwargs,
        )
        return target

    def open(self) -> zarr.Array:
        """Open the Zarr array for reading or writing and return it.

        Note that the Zarr array must have been created or this method will raise an exception.
        """
        # r+ means read/write, fail if it doesn't exist
        return open_backend_array(
            self.store,
            mode="r+",
            shape=self.shape,
            dtype=self.dtype,
            chunks=self.chunks,
            path=self.path,
            **self.kwargs,
        )

    def __repr__(self) -> str:
        return f"cubed.storage.zarr.LazyZarrArray<shape={self.shape}, dtype={self.dtype}, chunks={self.chunks}>"


T_ZarrArray = Union[zarr.Array, LazyZarrArray]


def lazy_zarr_array(
    store: T_Store,
    shape: T_Shape,
    dtype: T_DType,
    chunks: T_RegularChunks,
    path: Optional[str] = None,
    **kwargs,
) -> LazyZarrArray:
    return LazyZarrArray(
        store,
        shape,
        dtype,
        chunks,
        path=path,
        **kwargs,
    )


def open_if_lazy_zarr_array(array: T_ZarrArray) -> zarr.Array:
    """If array is a LazyZarrArray then open it, leaving other arrays unchanged."""
    return array.open() if isinstance(array, LazyZarrArray) else array



import math
import platform
import shutil
import sys
from functools import partial, reduce

import pandas as pd
import pytest

pytest.importorskip("memray")

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import partial_reduce
from cubed.core.optimization import multiple_inputs_optimize_dag
from cubed.diagnostics.history import HistoryCallback
from cubed.diagnostics.mem_warn import MemoryWarningCallback
from cubed.diagnostics.memray import MemrayCallback
from cubed.runtime.create import create_executor
from cubed.tests.test_core import sqrts
from cubed.tests.utils import LITHOPS_LOCAL_CONFIG

pd.set_option("display.max_columns", None)


ALLOWED_MEM = 2_000_000_000

EXECUTORS = {}

if platform.system() != "Windows":
    EXECUTORS["processes"] = create_executor("processes")

    # Run with max_tasks_per_child=1 so that each task is run in a new process,
    # allowing us to perform a stronger check on peak memory
    if sys.version_info >= (3, 11):
        executor_options = dict(max_tasks_per_child=1)
        EXECUTORS["processes-single-task"] = create_executor(
            "processes", executor_options
        )

try:
    executor_options = dict(config=LITHOPS_LOCAL_CONFIG, wait_dur_sec=0.1)
    EXECUTORS["lithops"] = create_executor("lithops", executor_options)
except ImportError:
    pass


@pytest.fixture
def spec(tmp_path, reserved_mem):
    return cubed.Spec(tmp_path, allowed_mem=ALLOWED_MEM, reserved_mem=reserved_mem)


@pytest.fixture(
    scope="module",
    params=EXECUTORS.values(),
    ids=EXECUTORS.keys(),
)
def executor(request):
    return request.param


@pytest.fixture(scope="module")
def reserved_mem(executor):
    res = cubed.measure_reserved_mem(executor) * 1.1  # add some wiggle room
    return round_up_to_multiple(res, 10_000_000)  # round up to nearest multiple of 10MB


def round_up_to_multiple(x, multiple=10):
    """Round up to the nearest multiple"""
    return math.ceil(x / multiple) * multiple


# Array Object


@pytest.mark.slow
def test_index(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[1:, :]
    run_operation(tmp_path, executor, "index", b)


@pytest.mark.slow
def test_index_chunk_aligned(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[0:5000, :]
    run_operation(tmp_path, executor, "index_chunk_aligned", b)


@pytest.mark.slow
def test_index_multiple_axes(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[1:, 1:]
    run_operation(tmp_path, executor, "index_multiple_axes", b)


@pytest.mark.slow
def test_index_step(tmp_path, spec, executor):
    # use 400MB chunks so that intermediate after indexing has 200MB chunks
    a = cubed.random.random(
        (20000, 10000), chunks=(10000, 5000), spec=spec
    )  # 400MB chunks
    b = a[::2, :]
    run_operation(tmp_path, executor, "index_step", b)


# Creation Functions


@pytest.mark.slow
def test_eye(tmp_path, spec, executor):
    a = xp.eye(10000, 10000, chunks=(5000, 5000), spec=spec)
    run_operation(tmp_path, executor, "eye", a)


@pytest.mark.slow
def test_tril(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.tril(a)
    run_operation(tmp_path, executor, "tril", b)


# Elementwise Functions


@pytest.mark.slow
@pytest.mark.parametrize("optimize_graph", [False, True])
def test_add(tmp_path, spec, executor, optimize_graph):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.add(a, b)
    run_operation(tmp_path, executor, "add", c, optimize_graph=optimize_graph)


@pytest.mark.slow
def test_add_reduce_left(tmp_path, spec, executor):
    # Perform the `add` operation repeatedly on pairs of arrays, also known as fold left.
    # See https://en.wikipedia.org/wiki/Fold_(higher-order_function)
    #
    # o   o
    #  \ /
    #   o   o
    #    \ /
    #     o   o
    #      \ /
    #       o
    #
    # Fusing fold left operations will result in a single fused operation.
    n_arrays = 10
    arrs = [
        cubed.random.random((10000, 10000), chunks=(5000, 5000), spec=spec)
        for _ in range(n_arrays)
    ]
    result = reduce(lambda x, y: xp.add(x, y), arrs)
    opt_fn = partial(multiple_inputs_optimize_dag, max_total_source_arrays=n_arrays * 2)
    run_operation(
        tmp_path, executor, "add_reduce_left", result, optimize_function=opt_fn
    )


@pytest.mark.slow
def test_add_reduce_right(tmp_path, spec, executor):
    # Perform the `add` operation repeatedly on pairs of arrays, also known as fold right.
    # See https://en.wikipedia.org/wiki/Fold_(higher-order_function)
    #
    #     o   o
    #      \ /
    #   o   o
    #    \ /
    # o   o
    #  \ /
    #   o
    #
    # Note that fusing fold right operations will result in unbounded memory usage unless care
    # is taken to limit fusion - which `multiple_inputs_optimize_dag` will do, with the result
    # that there is more than one fused operation (not a single fused operation).
    n_arrays = 10
    arrs = [
        cubed.random.random((10000, 10000), chunks=(5000, 5000), spec=spec)
        for _ in range(n_arrays)
    ]
    result = reduce(lambda x, y: xp.add(y, x), reversed(arrs))
    opt_fn = partial(multiple_inputs_optimize_dag, max_total_source_arrays=n_arrays * 2)
    run_operation(
        tmp_path, executor, "add_reduce_right", result, optimize_function=opt_fn
    )


@pytest.mark.slow
def test_negative(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.negative(a)
    run_operation(tmp_path, executor, "negative", b)


# Linear Algebra Functions


@pytest.mark.slow
def test_matmul(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.matmul(c, d)
    run_operation(tmp_path, executor, "matmul", e)


@pytest.mark.slow
def test_matrix_transpose(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.matrix_transpose(a)
    run_operation(tmp_path, executor, "matrix_transpose", b)


@pytest.mark.slow
def test_tensordot(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.tensordot(c, d, axes=1)
    run_operation(tmp_path, executor, "tensordot", e)


# Manipulation Functions


@pytest.mark.slow
def test_concat(tmp_path, spec, executor):
    # Note 'a' has one fewer element in axis=0 to force chunking to cross array boundaries
    a = cubed.random.random(
        (9999, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.concat((a, b), axis=0)
    run_operation(tmp_path, executor, "concat", c)


@pytest.mark.slow
def test_flip(tmp_path, spec, executor):
    # Note 'a' has one fewer element in axis=0 to force chunking to cross array boundaries
    a = cubed.random.random(
        (9999, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.flip(a, axis=0)
    run_operation(tmp_path, executor, "flip", b)


@pytest.mark.slow
def test_flip_multiple_axes(tmp_path, spec, executor):
    # Note 'a' has one fewer element in both axes to force chunking to cross array boundaries
    a = cubed.random.random(
        (9999, 9999), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.flip(a)
    run_operation(tmp_path, executor, "flip_multiple_axes", b)


@pytest.mark.slow
def test_repeat(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.repeat(a, 3, axis=0)
    run_operation(tmp_path, executor, "repeat", b)


@pytest.mark.slow
def test_reshape(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    # need intermediate reshape due to limitations in Dask's reshape_rechunk
    b = xp.reshape(a, (5000, 2, 10000))
    c = xp.reshape(b, (5000, 20000))
    run_operation(tmp_path, executor, "reshape", c)


@pytest.mark.slow
def test_stack(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.stack((a, b), axis=0)
    run_operation(tmp_path, executor, "stack", c)


@pytest.mark.slow
def test_unstack(tmp_path, spec, executor):
    a = cubed.random.random(
        (2, 10000, 10000), chunks=(2, 5000, 5000), spec=spec
    )  # 400MB chunks
    b, c = xp.unstack(a)
    run_operation(tmp_path, executor, "unstack", b, c)


# Searching Functions


@pytest.mark.slow
def test_argmax(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.argmax(a, axis=0)
    run_operation(tmp_path, executor, "argmax", b)


# Statistical Functions


@pytest.mark.slow
def test_cumulative_sum(tmp_path, spec, executor):
    a = cubed.random.random(
        (50000, 5000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.cumulative_sum(a, axis=0)
    run_operation(tmp_path, executor, "cumulative_sum", b)


@pytest.mark.slow
def test_max(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.max(a, axis=0)
    run_operation(tmp_path, executor, "max", b)


@pytest.mark.slow
def test_mean(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.mean(a, axis=0)
    run_operation(tmp_path, executor, "mean", b)


@pytest.mark.slow
def test_sum_partial_reduce(tmp_path, spec, executor):
    a = cubed.random.random(
        (40000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = partial_reduce(a, nxp.sum, split_every={0: 8})
    run_operation(tmp_path, executor, "sum_partial_reduce", b)


# Linear algebra extension


@pytest.mark.slow
def test_qr(tmp_path, spec, executor):
    a = cubed.random.random(
        (40000, 1000), chunks=(5000, 1000), spec=spec
    )  # 40MB chunks
    q, r = xp.linalg.qr(a)
    # don't optimize graph so we use as much memory as possible (reading from Zarr)
    run_operation(tmp_path, executor, "qr", q, r, optimize_graph=False)


# Multiple outputs


@pytest.mark.slow
def test_sqrts(tmp_path, spec, executor):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b, c = sqrts(a)
    # don't optimize graph so we use as much memory as possible (reading from Zarr)
    run_operation(tmp_path, executor, "sqrts", b, c, optimize_graph=False)


# Internal functions


def run_operation(
    tmp_path,
    executor,
    name,
    *results,
    optimize_graph=True,
    optimize_function=None,
):
    # cubed.visualize(
    #     *results, filename=f"cubed-{name}-unoptimized", optimize_graph=False, show_hidden=True
    # )
    # cubed.visualize(
    #     *results, filename=f"cubed-{name}", optimize_function=optimize_function
    # )
    hist = HistoryCallback()
    mem_warn = MemoryWarningCallback()
    memray = MemrayCallback(mem_threshold=30_000_000)
    # use None for each store to write to temporary zarr
    cubed.store(
        results,
        (None,) * len(results),
        executor=executor,
        callbacks=[hist, mem_warn, memray],
        optimize_graph=optimize_graph,
        optimize_function=optimize_function,
    )

    df = hist.stats_df
    print(df)

    # check peak memory does not exceed allowed mem
    assert (df["peak_measured_mem_end_mb_max"] <= ALLOWED_MEM // 1_000_000).all()

    # check change in peak memory is no more than projected mem
    assert (df["peak_measured_mem_delta_mb_max"] <= df["projected_mem_mb"]).all()

    # check memray peak memory allocated is no more than projected mem
    for op_name, stats in memray.stats.items():
        assert (
            stats.peak_memory_allocated
            <= df.query(f"name=='{op_name}'")["projected_mem_mb"].item() * 1_000_000
        ), f"projected mem exceeds memray's peak allocated for {op_name}"

    # check projected_mem_utilization does not exceed 1
    # except on processes executor that runs multiple tasks in a process
    if (
        executor.name != "processes"
        or executor.kwargs.get("max_tasks_per_child", None) == 1
    ):
        assert (df["projected_mem_utilization"] <= 1.0).all()

    # delete temp files for this test immediately since they are so large
    shutil.rmtree(tmp_path)



from __future__ import annotations

from numbers import Integral


def coerce_depth(ndim, depth):
    default = 0
    if depth is None:
        depth = default
    if isinstance(depth, Integral):
        depth = (depth,) * ndim
    if isinstance(depth, tuple):
        depth = dict(zip(range(ndim), depth))
    if isinstance(depth, dict):
        depth = {ax: depth.get(ax, default) for ax in range(ndim)}
    return coerce_depth_type(ndim, depth)


def coerce_depth_type(ndim, depth):
    for i in range(ndim):
        if isinstance(depth[i], tuple):
            depth[i] = tuple(int(d) for d in depth[i])
        else:
            depth[i] = int(depth[i])
    return depth


def coerce_boundary(ndim, boundary):
    default = "none"
    if boundary is None:
        boundary = default
    if not isinstance(boundary, (tuple, dict)):
        boundary = (boundary,) * ndim
    if isinstance(boundary, tuple):
        boundary = dict(zip(range(ndim), boundary))
    if isinstance(boundary, dict):
        boundary = {ax: boundary.get(ax, default) for ax in range(ndim)}
    return boundary



import asyncio
import contextlib
import os
from asyncio.exceptions import TimeoutError
from typing import Any, Callable, Optional, Sequence

import modal
from modal.exception import ConnectionError
from networkx import MultiDiGraph
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from cubed import config
from cubed.runtime.asyncio import async_map_dag
from cubed.runtime.backup import use_backups_default
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import asyncio_run, execute_with_stats
from cubed.spec import Spec, spec_from_config
from cubed.utils import convert_to_bytes

app = modal.App("cubed-app", include_source=True)

# Get Modal App settings from Cubed configuration. Note that we have to do this
# globally, since Modal remote functions have to be defined globally.
if modal.is_local():
    spec = spec_from_config(config)
    executor_options = spec.executor_options
else:
    executor_options = {}

runtime_memory = convert_to_bytes(executor_options.get("memory", "2GB"))
runtime_memory_mib = runtime_memory // 1_000_000
retries = executor_options.get("retries", 2)
timeout = executor_options.get("timeout", 180)
cloud = executor_options.get("cloud", "aws")
region = executor_options.get("region", None)
if modal.is_local() and region is None:
    raise ValueError(
        "Must set region when running using Modal, via the Cubed 'spec.executor_options.region' configuration setting."
    )

requirements_file = executor_options.get("requirements_file", None)

if requirements_file:
    image = modal.Image.debian_slim().pip_install_from_requirements(requirements_file)
else:
    if cloud == "aws":
        image = modal.Image.debian_slim().pip_install(
            [
                "array-api-compat",
                "donfig",
                "mypy_extensions",  # for rechunker
                "ndindex",
                "networkx",
                "obstore",
                "psutil",
                "pytest-mock",  # TODO: only needed for tests
                "tenacity",
                "toolz",
                "zarr",
            ]
        )
    elif cloud == "gcp":
        image = modal.Image.debian_slim().pip_install(
            [
                "array-api-compat",
                "donfig",
                "fsspec",
                "mypy_extensions",  # for rechunker
                "ndindex",
                "networkx",
                "psutil",
                "pytest-mock",  # TODO: only needed for tests
                "gcsfs",
                "tenacity",
                "toolz",
                "zarr",
            ]
        )
    else:
        raise ValueError(f"Unrecognized cloud: {cloud}")

secret_name = executor_options.get("secret", None)
if secret_name is None:
    if cloud == "aws":
        secret_name = "my-aws-secret"
    elif secret_name == "gcp":
        secrets_name = "my-googlecloud-secret"
    else:
        raise ValueError(f"Unrecognized cloud: {cloud}")
secrets = [modal.Secret.from_name(secret_name)]


def check_runtime_memory(spec):
    allowed_mem = spec.allowed_mem if spec is not None else None
    if allowed_mem is not None:
        if runtime_memory < allowed_mem:
            raise ValueError(
                f"Runtime memory ({runtime_memory}) is less than allowed_mem ({allowed_mem})"
            )


# We use a class so for GCP we can set up credentials by hooking into the container lifecycle
@app.cls(
    image=image,
    secrets=secrets,
    cpu=1.0,
    memory=runtime_memory_mib,  # modal memory is in MiB
    retries=retries,
    timeout=timeout,
    cloud=cloud,
    region=region,
)
class Container:
    @modal.enter()
    def set_up_credentials(self):
        if os.getenv("MODAL_CLOUD_PROVIDER") == "CLOUD_PROVIDER_GCP":
            json = os.environ["SERVICE_ACCOUNT_JSON"]
            path = os.path.abspath("application_credentials.json")
            with open(path, "w") as f:
                f.write(json)
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = path

    @modal.method()
    def run_remotely(self, input, func=None, config=None, name=None, compute_id=None):
        print(f"running remotely on {input} in {os.getenv('MODAL_REGION')}")
        # note we can't use the execution_stat decorator since it doesn't work with modal decorators
        result, stats = execute_with_stats(func, input, config=config)
        return result, stats


def modal_create_futures_func(function: Callable[..., Any]):
    def create_futures_func(input, **kwargs):
        return [
            (i, asyncio.ensure_future(function.remote.aio(i, **kwargs))) for i in input
        ]

    return create_futures_func


class ModalExecutor(DagExecutor):
    """An execution engine that uses Modal's async API."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "modal"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        # remove executor options as they should already have been used in defining the remote functions
        for executor_option in (
            "memory",
            "retries",
            "timeout",
            "cloud",
            "region",
            "requirements_file",
            "secret",
        ):
            merged_kwargs.pop(executor_option, None)
        asyncio_run(
            self._async_execute_dag(
                dag,
                callbacks=callbacks,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )

    # This just retries the initial connection attempt, not the function calls
    @retry(
        reraise=True,
        retry=retry_if_exception_type((TimeoutError, ConnectionError)),
        stop=stop_after_attempt(3),
    )
    async def _async_execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        cloud: Optional[str] = None,
        compute_arrays_in_parallel: Optional[bool] = None,
        **kwargs,
    ) -> None:
        if spec is not None:
            check_runtime_memory(spec)
            if "use_backups" not in kwargs and use_backups_default(spec):
                kwargs["use_backups"] = True

        function = Container().run_remotely

        enable_output = kwargs.pop("enable_output", False)
        cm = modal.enable_output() if enable_output else contextlib.nullcontext()

        with cm:
            async with app.run():
                create_futures_func = modal_create_futures_func(function)
                await async_map_dag(
                    create_futures_func,
                    dag=dag,
                    callbacks=callbacks,
                    compute_arrays_in_parallel=compute_arrays_in_parallel,
                    **kwargs,
                )



import asyncio
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import nullcontext
from functools import partial
from itertools import islice
from pathlib import Path

from cubed.runtime.types import OperationStartEvent, TaskEndEvent
from cubed.utils import peak_measured_mem

try:
    import memray
except ImportError:
    memray = None  # type: ignore

sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def execute_with_stats(function, *args, **kwargs):
    """Invoke function and measure timing information and peak memory usage.

    Returns the result of the function call and a stats dictionary.
    """

    peak_measured_mem_start = peak_measured_mem()
    function_start_tstamp = time.time()
    result = function(*args, **kwargs)
    function_end_tstamp = time.time()
    peak_measured_mem_end = peak_measured_mem()
    return result, dict(
        function_start_tstamp=function_start_tstamp,
        function_end_tstamp=function_end_tstamp,
        peak_measured_mem_start=peak_measured_mem_start,
        peak_measured_mem_end=peak_measured_mem_end,
    )


def execute_with_timing(function, *args, **kwargs):
    """Invoke function and measure timing information.

    Returns the result of the function call and a stats dictionary.
    """

    function_start_tstamp = time.time()
    result = function(*args, **kwargs)
    function_end_tstamp = time.time()
    return result, dict(
        function_start_tstamp=function_start_tstamp,
        function_end_tstamp=function_end_tstamp,
    )


def execution_stats(func):
    """Decorator to measure timing information and peak memory usage of a function call."""

    return partial(execute_with_stats, func)


def execution_timing(func):
    """Decorator to measure timing information of a function call."""

    return partial(execute_with_timing, func)


def execute_with_memray(function, input, **kwargs):
    # only run memray if installed, and only for first input (for operations that run on block locations)
    if (
        memray is not None
        and "compute_id" in kwargs
        and isinstance(input, list)
        and all(isinstance(i, int) for i in input)
        and sum(input) == 0
    ):
        compute_id = kwargs["compute_id"]
        name = kwargs["name"]
        memray_dir = Path(f"history/{compute_id}/memray")
        memray_dir.mkdir(parents=True, exist_ok=True)
        cm = memray.Tracker(memray_dir / f"{name}.bin")
    else:
        cm = nullcontext()
    with cm:
        result = result = function(input, **kwargs)
        return result


def profile_memray(func):
    """Decorator to profile a function call with memray."""
    return partial(execute_with_memray, func)


def handle_operation_start_callbacks(callbacks, name):
    if callbacks is not None:
        event = OperationStartEvent(name)
        [callback.on_operation_start(event) for callback in callbacks]


def handle_callbacks(callbacks, result, stats):
    """Construct a TaskEndEvent from stats and send to all callbacks."""

    if callbacks is not None:
        if "task_result_tstamp" not in stats:
            task_result_tstamp = time.time()
            event = TaskEndEvent(
                result=result,
                task_result_tstamp=task_result_tstamp,
                **stats,
            )
        else:
            event = TaskEndEvent(result=result, **stats)
        [callback.on_task_end(event) for callback in callbacks]


# Like asyncio.run(), but works in a Jupyter notebook
# Based on https://stackoverflow.com/a/75341431
def asyncio_run(coro):
    try:
        asyncio.get_running_loop()  # Triggers RuntimeError if no running event loop
    except RuntimeError:
        return asyncio.run(coro)
    else:
        # Create a separate thread so we can block before returning
        with ThreadPoolExecutor(1) as pool:
            return pool.submit(lambda: asyncio.run(coro)).result()


# this will be in Python 3.12 https://docs.python.org/3.12/library/itertools.html#itertools.batched
def batched(iterable, n):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



import functools
import importlib.util

import numpy as np
import numpy.testing as npt

import cubed


@functools.cache
def has_cupy() -> bool:
    return importlib.util.find_spec("cupy") is not None


@functools.singledispatch
def to_numpy(a):
    return np.asarray(a)


@to_numpy.register(cubed.Array)
def _(a: cubed.Array) -> np.ndarray:
    return to_numpy(a.compute())


if has_cupy():
    import cupy

    @to_numpy.register(cupy.ndarray)
    def _(a):
        return a.get()


def assert_array_equal(a, b, **kwargs):
    a = to_numpy(a)
    b = to_numpy(b)
    npt.assert_array_equal(a, b, **kwargs)


def assert_allclose(a, b, **kwargs):
    a = to_numpy(a)
    b = to_numpy(b)
    npt.assert_allclose(a, b, **kwargs)



import itertools
import math
from math import ceil, prod
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple

import numpy as np

from cubed.primitive.types import CubedArrayProxy, CubedCopySpec, PrimitiveOperation
from cubed.runtime.types import CubedPipeline
from cubed.storage.zarr import T_ZarrArray, lazy_zarr_array
from cubed.types import T_RegularChunks, T_Shape, T_Store
from cubed.utils import itemsize
from cubed.vendor.rechunker.algorithm import rechunking_plan

sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def rechunk(
    source: T_ZarrArray,
    source_array_name: str,
    int_array_name: str,
    target_array_name: str,
    target_chunks: T_RegularChunks,
    allowed_mem: int,
    reserved_mem: int,
    target_store: T_Store,
    storage_options: Optional[Dict[str, Any]] = None,
) -> List[PrimitiveOperation]:
    """Change the chunking of an array, without changing its shape or dtype.

    Parameters
    ----------
    source : Zarr array
    target_chunks : tuple
        The desired chunks of the array after rechunking.
    allowed_mem : int
        The memory available to a worker for running a task, in bytes. Includes ``reserved_mem``.
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes
    target_store : str
        Path to output Zarr store.

    Returns
    -------
    PrimitiveOperation to run the operation
    """

    # rechunker doesn't take account of uncompressed and compressed copies of the
    # input and output array chunk/selection, so adjust appropriately
    rechunker_max_mem = (allowed_mem - reserved_mem) // 4

    # we assume that rechunker is going to use all the memory it is allowed to
    projected_mem = allowed_mem

    read_proxy, int_proxy, write_proxy = _setup_array_rechunk(
        source_array=source,
        int_array_name=int_array_name,
        target_array_name=target_array_name,
        target_chunks=target_chunks,
        max_mem=rechunker_max_mem,
        target_store=target_store,
        storage_options=storage_options,
    )

    target = write_proxy.array

    if int_proxy is None:
        copy_spec = CubedCopySpec(read_proxy, write_proxy)
        num_tasks = total_chunks(write_proxy.array.shape, write_proxy.chunks)
        return [
            spec_to_primitive_op(
                copy_spec,
                source_array_name,
                target,
                projected_mem,
                allowed_mem,
                reserved_mem,
                num_tasks,
            )
        ]

    else:
        # break spec into two if there's an intermediate
        intermediate = int_proxy.array
        copy_spec1 = CubedCopySpec(read_proxy, int_proxy)
        num_tasks = total_chunks(copy_spec1.write.array.shape, copy_spec1.write.chunks)
        op1 = spec_to_primitive_op(
            copy_spec1,
            source_array_name,
            intermediate,
            projected_mem,
            allowed_mem,
            reserved_mem,
            num_tasks,
        )

        copy_spec2 = CubedCopySpec(int_proxy, write_proxy)
        num_tasks = total_chunks(copy_spec2.write.array.shape, copy_spec2.write.chunks)
        op2 = spec_to_primitive_op(
            copy_spec2,
            int_array_name,
            target,
            projected_mem,
            allowed_mem,
            reserved_mem,
            num_tasks,
        )

        return [op1, op2]


# from rechunker, but simpler since it only has to handle Zarr arrays
def _setup_array_rechunk(
    source_array: T_ZarrArray,
    int_array_name: str,
    target_array_name: str,
    target_chunks: T_RegularChunks,
    max_mem: int,
    target_store: T_Store,
    storage_options: Optional[Dict[str, Any]] = None,
) -> Tuple[CubedArrayProxy, Optional[CubedArrayProxy], CubedArrayProxy]:
    shape = source_array.shape
    source_chunks = source_array.chunks
    dtype = source_array.dtype

    read_chunks, int_chunks, write_chunks = rechunking_plan(
        shape,
        source_chunks,
        target_chunks,
        itemsize(dtype),
        max_mem,
    )

    # create target
    shape = tuple(int(x) for x in shape)  # ensure python ints for serialization
    target_chunks = tuple(int(x) for x in target_chunks)
    int_chunks = tuple(int(x) for x in int_chunks)
    write_chunks = tuple(int(x) for x in write_chunks)

    target_array = lazy_zarr_array(
        target_store,
        shape,
        dtype,
        chunks=target_chunks,
        path=target_array_name,  # use array name for path within store
        storage_options=storage_options,
    )

    if read_chunks == write_chunks:
        int_array = None
    else:
        # do intermediate store
        int_array = lazy_zarr_array(
            target_store,
            shape,
            dtype,
            chunks=int_chunks,
            path=int_array_name,  # use array name for path within store
            storage_options=storage_options,
        )

    read_proxy = CubedArrayProxy(source_array, read_chunks)
    int_proxy = (
        CubedArrayProxy(int_array, int_chunks) if int_array is not None else None
    )
    write_proxy = CubedArrayProxy(target_array, write_chunks)
    return read_proxy, int_proxy, write_proxy


def total_chunks(shape: T_Shape, chunks: T_RegularChunks) -> int:
    # cf rechunker's chunk_keys
    return prod(ceil(s / c) for s, c in zip(shape, chunks))


# differs from rechunker's chunk_keys to return a list rather than a tuple, to keep lithops happy
def chunk_keys(
    shape: Tuple[int, ...], chunks: Tuple[int, ...]
) -> Iterator[List[slice]]:
    """Iterator over array indexing keys of the desired chunk sized.

    The union of all keys indexes every element of an array of shape ``shape``
    exactly once. Each array resulting from indexing is of shape ``chunks``,
    except possibly for the last arrays along each dimension (if ``chunks``
    do not even divide ``shape``).
    """
    ranges = [range(math.ceil(s / c)) for s, c in zip(shape, chunks)]
    for indices in itertools.product(*ranges):
        yield [
            slice(c * i, min(c * (i + 1), s)) for i, s, c in zip(indices, shape, chunks)
        ]


# need a ChunkKeys Iterable instead of a generator, to avoid beam pickle error
class ChunkKeys(Iterable[Tuple[slice, ...]]):
    def __init__(self, shape: Tuple[int, ...], chunks: Tuple[int, ...]):
        self.shape = shape
        self.chunks = chunks

    def __iter__(self):
        return chunk_keys(self.shape, self.chunks)


def copy_read_to_write(chunk_key: Tuple[slice], *, config: CubedCopySpec) -> None:
    # workaround limitation of lithops.utils.verify_args
    if isinstance(chunk_key, list):
        chunk_key = tuple(chunk_key)
    data = np.asarray(config.read.open()[chunk_key])
    config.write.open()[chunk_key] = data


def spec_to_primitive_op(
    spec: CubedCopySpec,
    source_array_name: str,
    target_array: Any,
    projected_mem: int,
    allowed_mem: int,
    reserved_mem: int,
    num_tasks: int,
) -> PrimitiveOperation:
    # typing won't work until we start using numpy types
    shape = spec.read.array.shape  # type: ignore
    pipeline = CubedPipeline(
        copy_read_to_write,
        gensym("copy_read_to_write"),
        ChunkKeys(shape, spec.write.chunks),
        spec,
    )
    source_array_names = [source_array_name]
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=source_array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable_with_predecessors=False,
        write_chunks=spec.write.chunks,
    )
